import os
import pandas as pd
import numpy as np
from datetime import datetime
from loone.data.model_variables import M_var as MVarClass
from loone.utils import load_config, lonino_functions
from loone.data import Data as DClass


# I determine daily values for the Tributary conditions and Seasonal/Multi-Seasonal LONINO classes
# using a weekly Trib. Condition data and Monthly LONINO data.
def Trib_HC(workspace: str):
    """
    This function generates daily values for the Tributary conditions and Seasonal/Multi-Seasonal LONINO classes
    using a weekly Trib. Condition data and Monthly LONINO data.

    It generates the Tributary Condition Dataframe, Trib_Cond_df, and the LONINO Dataframe, LONINO_df.
    It uses the Trib_Cond_df to calculate the Tributary Condition Index (TCI) and the LONINO_df to calculate
    the LONINO seasonal and multi-seasonal classes.

    Finally, it generates a daily date range, TC_LONINO_df, and assigns the TCI, LONINO seasonal, and LONINO
    multi-seasonal classes to this dataframe.

    Returns:
        TC_LONINO_df (pd.DataFrame): A dataframe with the daily values of the Tributary conditions and Seasonal/Multi-Seasonal LONINO classes.

    """
    os.chdir(workspace)
    config = load_config(workspace)
    Data = DClass(workspace)
    M_var = MVarClass(config)
    # Generate weekly time step date column where frequency is 'W-Fri' to start on 01/01/2008.
    # FIXME: Always check here for start date, end date, and frequency to match with the Trib. Condition weekly data obtained.
    year, month, day = map(int, config["start_date_entry"])
    startdate = datetime(year, month, day).date()
    year, month, day = map(int, config["end_date_entry"])
    enddate = datetime(year, month, day).date()
    year, month, day = map(int, config["end_date_tc"])
    enddate_TC = datetime(year, month, day).date()
    date_rng_3 = pd.date_range(start=startdate, end=enddate_TC, freq="W-Fri")
    # Generate the Tributary Condition Dataframe.
    Trib_Cond_df = pd.DataFrame(date_rng_3, columns=["date"])
    TC_Count = len(Trib_Cond_df.index)
    for i in range(TC_Count):
        M_var.RF_Cls[i] = lonino_functions.RF_Cls(
            Data.Wkly_Trib_Cond["NetRF"].iloc[i]
        )
        M_var.MainTrib_Cls[i] = lonino_functions.MainTrib_Cls(
            Data.Wkly_Trib_Cond["S65E"].iloc[i]
        )
        M_var.Palmer_Cls[i] = lonino_functions.Palmer_Cls(
            Data.Wkly_Trib_Cond["Palmer"].iloc[i]
        )
        M_var.NetInflow_Cls[i] = lonino_functions.NetInflow_Cls(
            Data.Wkly_Trib_Cond["NetInf"].iloc[i]
        )
        M_var.Max_RF_MainTrib[i] = max(M_var.RF_Cls[i], M_var.MainTrib_Cls[i])
        M_var.Max_Palmer_NetInf[i] = max(
            M_var.Palmer_Cls[i], M_var.NetInflow_Cls[i]
        )
    if config["tci"] == 1:  # Tributary Condition Index
        Trib_Cond_df["TCI"] = M_var.Max_Palmer_NetInf
    else:
        Trib_Cond_df["TCI"] = M_var.Max_RF_MainTrib
    # Generate a monthly time step date column
    date_rng_4 = pd.date_range(start=startdate, end=enddate, freq="MS")
    # Create a LONINO Dataframe
    LONINO_df = pd.DataFrame(date_rng_4, columns=["date"])
    LONINO_Count = len(LONINO_df.index)
    for i in range(LONINO_Count):
        M_var.Seas[i] = Data.LONINO_Seas_data[
            str(LONINO_df["date"].iloc[i].month)
        ].iloc[LONINO_df["date"].iloc[i].year - config["start_year"]]
        M_var.M_Seas[i] = Data.LONINO_Mult_Seas_data[
            str(LONINO_df["date"].iloc[i].month)
        ].iloc[LONINO_df["date"].iloc[i].year - config["start_year"]]
    LONINO_df["LONINO_Seas"] = M_var.Seas
    LONINO_df["LONINO_Mult_Seas"] = M_var.M_Seas
    for i in range(config["month_n"]):
        M_var.LONINO_Seas_cls[i] = lonino_functions.LONINO_Seas_cls(
            LONINO_df["LONINO_Seas"].iloc[i]
        )
        M_var.LONINO_M_Seas_cls[i] = lonino_functions.LONINO_M_Seas_cls(
            LONINO_df["LONINO_Mult_Seas"].iloc[i]
        )
    LONINO_df["LONINO_Seasonal_Cls"] = M_var.LONINO_Seas_cls
    LONINO_df["LONINO_Mult_Seasonal_Cls"] = M_var.LONINO_M_Seas_cls
    # Generate a daily date range
    date_rng_5 = pd.date_range(start=startdate, end=enddate, freq="D")
    TC_LONINO_df = pd.DataFrame(date_rng_5, columns=["Date"])
    row_nm = len(TC_LONINO_df.index)
    Trib_Cond = np.zeros(row_nm)
    for i in range(row_nm):
        Trib_Cond[i] = Trib_Cond_df["TCI"].iloc[int(i / 7)]
    TC_LONINO_df["Tributary_Condition"] = Trib_Cond
    data_S_MS = [
        LONINO_df["date"],
        LONINO_df["LONINO_Seasonal_Cls"],
        LONINO_df["LONINO_Mult_Seasonal_Cls"],
    ]
    headers_S_MS = ["date", "LONINO_Seasonal_Class", "LONINO_MSeasonal_Class"]
    LONINO_Seas_MSeas_df = pd.concat(data_S_MS, axis=1, keys=headers_S_MS)
    LONINO_Seas_MSeas_df = LONINO_Seas_MSeas_df.set_index("date")
    LONINO_Seas_MSeas_df = LONINO_Seas_MSeas_df.reindex(
        date_rng_5, method="ffill"
    )
    LONINO_Seas_MSeas_df = LONINO_Seas_MSeas_df.reset_index()
    TC_LONINO_df["LONINO_Seasonal_Classes"] = LONINO_Seas_MSeas_df[
        "LONINO_Seasonal_Class"
    ]
    TC_LONINO_df["LONINO_MultiSeasonal_Classes"] = LONINO_Seas_MSeas_df[
        "LONINO_MSeasonal_Class"
    ]
    return TC_LONINO_df



import os
import argparse
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
from calendar import monthrange
from typing import List
from loone.utils import (
    load_config,
    replicate,
    df_wsms,
    stg_sto_ar,
    thc_class,
    lo_functions,
    dec_tree_functions,
    trib_hc,
)
from loone.data.model_variables import M_var as MVarClass
from loone.utils.wca_stages_class import WCA_Stages_Cls
from loone.data import Data as DClass


def _calculate_wsm_zone(i: int, model_variables: object, lo_functions: object, df_WSMs: pd.DataFrame) -> None:
    """
    Calculate the WSM zone for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        df_WSMs (pd.DataFrame): The WSMs DataFrame.

    Returns:
        None
    """
    model_variables.WSM_Zone[i + 2] = lo_functions.WSM_Zone(
        model_variables.Lake_Stage[i + 1],
        df_WSMs.at[i + 1, "WSM4"],
        df_WSMs.at[i + 1, "WSM3"],
        df_WSMs.at[i + 1, "WSM2"],
        df_WSMs.at[i + 1, "WSM1"],
    )


def _calculate_max_supply(i: int, model_variables: object, lo_functions: object, water_demand: pd.DataFrame,
                          config: dict) -> None:
    """
    Calculate the maximum supply for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        water_demand (pd.DataFrame): The water demand DataFrame.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Max_Supply[i + 2] = lo_functions.Max_Supply(
        model_variables.WSM_Zone[i + 2],
        water_demand.at[i, "Daily_demand"],
        config["z1_cutback"],
        config["z2_cutback"],
        config["z3_cutback"],
        config["z4_cutback"],
    )


def _calculate_losa_supply(i: int, model_variables: object, lo_functions: object, lo_model: pd.DataFrame,
                           config: dict) -> None:
    """
    Calculate the LOSA supply for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        lo_model (pd.DataFrame): The model DataFrame.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.LOSA_Supply[i + 2] = lo_functions.LOSA_Supply(
        model_variables.WSM_Zone[i + 2],
        lo_model.at[i + 2, "LOSA_dmd_SFWMM"],
        model_variables.Max_Supply[i + 2],
        config["opt_losa_ws"],
    )


def _calculate_ni_supply(i: int, model_variables: object, lo_model: pd.DataFrame) -> None:
    """
    Calculate the NI supply for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_model (pd.DataFrame): The model DataFrame.

    Returns:
        None
    """
    model_variables.NI_Supply[i + 2] = (
        lo_model.at[i + 2, "Net_Inflow"] - model_variables.LOSA_Supply[i + 2]
    )


def _calculate_cutback(i: int, model_variables: object, lo_model: pd.DataFrame) -> None:
    """
    Calculate the cutback for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_model (pd.DataFrame): The model DataFrame.

    Returns:
        None
    """
    cutback = lo_model.at[i + 2, "LOSA_dmd_SFWMM"] - model_variables.LOSA_Supply[i + 2]
    model_variables.Cut_back[i + 2] = cutback


def _calculate_demand_not_supplied(i: int, model_variables: object, lo_model: pd.DataFrame) -> None:
    """
    Calculate the demand not supplied for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_model (pd.DataFrame): The model DataFrame.

    Returns:
        None
    """
    if lo_model.at[i + 2, "LOSA_dmd_SFWMM"] == 0:
        demand_not_supplied = 0
    else:
        demand_not_supplied = (
            model_variables.Cut_back[i + 2] / lo_model.at[i + 2, "LOSA_dmd_SFWMM"]
        ) * 100
    model_variables.Dem_N_Sup[i + 2] = demand_not_supplied


def _calculate_initial_zone_code_and_lo_zone(model_variables: object, lo_functions: object,
                                             df_WSMs: pd.DataFrame) -> None:
    """
    Calculate the Zone_Code and LO_Zone for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        df_WSMs (pd.DataFrame): The WSMs DataFrame.

    Returns:
        None
    """
    model_variables.Zone_Code[0] = lo_functions.Zone_Code(
        model_variables.Lake_Stage[0],
        df_WSMs["A"].iloc[0],
        df_WSMs["B"].iloc[0],
        df_WSMs["C"].iloc[0],
        df_WSMs["D3"].iloc[0],
        df_WSMs["D2"].iloc[0],
        df_WSMs["D1"].iloc[0],
        df_WSMs["D0"].iloc[0],
        df_WSMs["WSM1"].iloc[0],
    )
    model_variables.LO_Zone[0] = lo_functions.LO_Zone(model_variables.Zone_Code[0])


def _calculate_zone_code(i: int, model_variables: object, lo_functions: object, df_WSMs: pd.DataFrame) -> None:
    """
    Calculate the Zone_Code for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        df_WSMs (pd.DataFrame): The WSMs DataFrame.

    Returns:
        None
    """
    model_variables.Zone_Code[i + 1] = lo_functions.Zone_Code(
        model_variables.Lake_Stage[i + 1],
        df_WSMs.at[i + 1, "A"],
        df_WSMs.at[i + 1, "B"],
        df_WSMs.at[i + 1, "C"],
        df_WSMs.at[i + 1, "D3"],
        df_WSMs.at[i + 1, "D2"],
        df_WSMs.at[i + 1, "D1"],
        df_WSMs.at[i + 1, "D0"],
        df_WSMs.at[i + 1, "WSM1"],
    )


def _generate_zone_column(i: int, model_variables: object, lo_functions: object) -> None:
    """
    Generate the Zone columns for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.

    Returns:
        None
    """
    model_variables.LO_Zone[i + 1] = lo_functions.LO_Zone(model_variables.Zone_Code[i + 1])


def _calculate_decision_tree_variables(i: int, model_variables: object, dec_tree_functions: object,
                                       tc_lonino_df: pd.DataFrame, config: dict, seasons: pd.DataFrame,
                                       dec_tree_df: pd.DataFrame, df_WSMs: pd.DataFrame) -> None:
    """
    Calculate the decision tree variables for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        dec_tree_functions (object): The decision tree functions object.
        tc_lonino_df (pd.DataFrame): The TC LONINO DataFrame.
        config (dict): The configuration dictionary.
        seasons (pd.DataFrame): The seasons DataFrame.
        dec_tree_df (pd.DataFrame): The decision tree DataFrame.
        df_WSMs (pd.DataFrame): The WSMs DataFrame.

    Returns:
        None
    """
    model_variables.Zone_D_Trib[i] = dec_tree_functions.Zone_D_Trib(
        tc_lonino_df.at[i, "Tributary_Condition"],
        config["opt_new_tree"],
    )
    model_variables.Zone_D_stage[i] = dec_tree_functions.Zone_D_stage(
        model_variables.Lake_Stage[i + 1], df_WSMs.at[i, "C-b"]
    )
    model_variables.Zone_D_Seas[i] = dec_tree_functions.Zone_D_Seas(
        tc_lonino_df.at[i, "LONINO_Seasonal_Classes"],
        model_variables.Zone_D_Trib[i],
        config["opt_new_tree"],
    )
    model_variables.Zone_D_MSeas[i] = dec_tree_functions.Zone_D_MSeas(
        tc_lonino_df.at[i, "LONINO_MultiSeasonal_Classes"]
    )
    model_variables.Zone_D_Branch_Code[i] = (
        model_variables.Zone_D_Trib[i] * 1000
        + model_variables.Zone_D_stage[i] * 100
        + model_variables.Zone_D_Seas[i] * 10
        + model_variables.Zone_D_MSeas[i] * 1
    )
    model_variables.Zone_D_Rel_Code[i] = dec_tree_functions.Zone_D_Rel_Code(
        model_variables.Zone_D_Branch_Code[i], config["opt_dec_tree"]
    )
    model_variables.Zone_C_Trib[i] = dec_tree_functions.Zone_C_Trib(
        tc_lonino_df.at[i, "Tributary_Condition"],
        config["opt_new_tree"],
    )
    model_variables.Zone_C_Seas[i] = dec_tree_functions.Zone_C_Seas(
        tc_lonino_df.at[i, "LONINO_Seasonal_Classes"],
        config["opt_new_tree"],
    )
    model_variables.Zone_C_MSeas[i] = dec_tree_functions.Zone_C_MSeas(
        tc_lonino_df.at[i, "LONINO_MultiSeasonal_Classes"]
    )
    model_variables.Zone_C_MetFcast[i] = dec_tree_functions.Zone_C_MetFcast(
        model_variables.Zone_C_Seas[i],
        tc_lonino_df.at[i, "LONINO_Seasonal_Classes"],
        config["zone_c_met_fcast_indicator"],
    )
    model_variables.Zone_C_Branch_Code[i] = (
        model_variables.Zone_C_Trib[i] * 1000
        + model_variables.Zone_C_MetFcast[i] * 100
        + model_variables.Zone_C_Seas[i] * 10
        + model_variables.Zone_C_MSeas[i] * 1
    )
    model_variables.Zone_C_Rel_Code[i] = dec_tree_functions.Zone_C_Rel_Code(
        model_variables.Zone_C_Branch_Code[i], config["opt_dec_tree"]
    )
    model_variables.Zone_B_Trib[i] = dec_tree_functions.Zone_B_Trib(
        tc_lonino_df.at[i, "Tributary_Condition"],
        config["opt_new_tree"],
    )
    model_variables.Zone_B_Stage[i] = dec_tree_functions.Zone_B_Stage(
        model_variables.Lake_Stage[i + 1], seasons.at[i, "Season"]
    )
    model_variables.Zone_B_Seas[i] = dec_tree_functions.Zone_B_Seas(
        tc_lonino_df.at[i, "LONINO_Seasonal_Classes"]
    )
    model_variables.Zone_B_Branch_Code[i] = (
        model_variables.Zone_B_Trib[i] * 1000
        + model_variables.Zone_B_Stage[i] * 100
        + dec_tree_df.at[i, "Zone_B_MetFcast"] * 10
        + model_variables.Zone_B_Seas[i] * 1
    )
    model_variables.Zone_B_Rel_Code[i] = dec_tree_functions.Zone_B_Rel_Code(
        model_variables.Zone_B_Branch_Code[i], config["opt_dec_tree"]
    )


def _calculate_release_levels(i: int, startdate: datetime, lo_model: pd.DataFrame, model_variables: object,
                              lo_functions: object, config: dict, tc_lonino_df: pd.DataFrame) -> None:
    """
    Calculate the release levels for the given index.

    Args:
        i (int): The current index.
        startdate (datetime): The start date.
        lo_model (pd.DataFrame): The model DataFrame.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        config (dict): The configuration dictionary.
        tc_lonino_df (pd.DataFrame): The TC LONINO DataFrame.

    Returns:
        None
    """
    model_variables.DecTree_Relslevel[i + 2] = lo_functions.DecTree_Relslevel(
        model_variables.Zone_Code[i + 1],
        model_variables.Zone_D_Rel_Code[i],
        model_variables.Zone_C_Rel_Code[i],
        model_variables.Zone_B_Rel_Code[i],
    )
    if i >= 3:
        if (
            startdate.month == lo_model.at[i, "date"].month
            and startdate.day == lo_model.at[i, "date"].day
            and (
                config["cs_flag"] == 0
                or startdate.year == lo_model.at[i, "date"].year
            )
        ):
            X2 = "SimDay1"
        else:
            X2 = lo_model.at[i, "date"].date()
        model_variables.DayFlags[i] = X2
    model_variables.PlsDay[i + 2] = lo_functions.PlsDay(
        model_variables.DayFlags[i + 2],
        model_variables.DecTree_Relslevel[i + 2],
        config["pls_day_switch"],
    )
    model_variables.Release_Level[i + 2] = lo_functions.release_level(
        model_variables.Release_Level[i + 1],
        model_variables.Lake_Stage[i + 1],
        tc_lonino_df.at[i, "Tributary_Condition"],
        model_variables.PlsDay[i + 2],
        model_variables.Zone_Code[i + 1],
        model_variables.DecTree_Relslevel[i + 2],
        config["max_qstg_trigger"],
    )
    if i >= 6:
        dh = model_variables.Lake_Stage[i + 1] - model_variables.Lake_Stage[i - 6]
        model_variables.dh_7days[i + 1] = dh


def _calculate_zone_codes(i: int, model_variables: object, lo_functions: object, df_WSMs: pd.DataFrame) -> None:
    """
    Calculate the zone codes for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        df_WSMs (pd.DataFrame): The WSMs DataFrame.

    Returns:
        None
    """
    model_variables.ZoneCodeminus1Code[i + 1] = lo_functions.ZoneCodeminus1Code(
        model_variables.Zone_Code[i + 1],
        df_WSMs.at[i + 1, "WSM1"],
        df_WSMs.at[i + 1, "D0"],
        df_WSMs.at[i + 1, "D1"],
        df_WSMs.at[i + 1, "D2"],
        df_WSMs.at[i + 1, "D3"],
        df_WSMs.at[i + 1, "C"],
        df_WSMs.at[i + 1, "B"],
        df_WSMs.at[i + 1, "A"],
    )
    model_variables.ZoneCodeCode[i + 1] = lo_functions.ZoneCodeCode(
        model_variables.Zone_Code[i + 1],
        df_WSMs.at[i + 1, "WSM1"],
        df_WSMs.at[i + 1, "D0"],
        df_WSMs.at[i + 1, "D1"],
        df_WSMs.at[i + 1, "D2"],
        df_WSMs.at[i + 1, "D3"],
        df_WSMs.at[i + 1, "C"],
        df_WSMs.at[i + 1, "B"],
        df_WSMs.at[i + 1, "A"],
    )
    model_variables.Fraction_of_Zone_height[i + 1] = (
        lo_functions.Fraction_of_Zone_height(
            model_variables.Zone_Code[i + 1],
            model_variables.Lake_Stage[i + 1],
            model_variables.ZoneCodeminus1Code[i + 1],
            model_variables.ZoneCodeCode[i + 1],
        )
    )


def _calculate_release_level_codes(i: int, model_variables: object, lo_functions: object, config: dict) -> None:
    """
    Calculate the release level codes for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.ReLevelCode_1[i + 2] = lo_functions.ReLevelCode_1(
        model_variables.Release_Level[i + 2],
        config["dstar_d1"],
        config["dstar_d2"],
        config["dstar_d3"],
        config["dstar_c"],
        config["dstar_b"],
    )
    model_variables.ReLevelCode_2[i + 2] = lo_functions.ReLevelCode_2(
        model_variables.Release_Level[i + 2],
        config["astar_d1"],
        config["astar_d2"],
        config["astar_d3"],
        config["astar_c"],
        config["astar_b"],
    )
    model_variables.ReLevelCode_3_S80[i + 2] = lo_functions.ReLevelCode_3_S80(
        model_variables.Release_Level[i + 2],
        config["bstar_s80_d1"],
        config["bstar_s80_d2"],
        config["bstar_s80_d3"],
        config["bstar_s80_c"],
        config["bstar_s80_b"],
    )


def _calculate_relevel_code_3_s77(i: int, model_variables: object, lo_functions: object, config: dict) -> None:
    """
    Calculate the ReLevelCode_3_S77 for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.ReLevelCode_3_S77[i + 2] = lo_functions.ReLevelCode_3_S77(
        model_variables.Release_Level[i + 2],
        config["bstar_s77_d1"],
        config["bstar_s77_d2"],
        config["bstar_s77_d3"],
        config["bstar_s77_c"],
        config["bstar_s77_b"],
    )


def _calculate_outlet2ds_mult(i: int, lo_model: pd.DataFrame, model_variables: object, lo_functions: object,
                              seasons: pd.DataFrame, config: dict) -> None:
    """
    Calculate the Outlet2DS multiplier for the given index.

    Args:
        i (int): The current index.
        lo_model (pd.DataFrame): The model DataFrame.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        seasons (pd.DataFrame): The seasons DataFrame.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Outlet2DS_Mult[i + 2] = lo_functions.Outlet2DS_Mult(
        seasons.at[i, "Season"],
        seasons.at[i, "Month"],
        model_variables.dh_7days[i + 1],
        model_variables.ReLevelCode_1[i + 2],
        model_variables.Fraction_of_Zone_height[i + 1],
        model_variables.ReLevelCode_2[i + 2],
        model_variables.ReLevelCode_3_S80[i + 2],
        config["opt_qreg_mult"],
    )
    model_variables.Outlet2DS_Mult_2[i + 2] = lo_functions.Outlet2DS_Mult_2(
        lo_model.at[i + 2, "date"].month,
        lo_model.at[i + 2, "date"].day,
        model_variables.PlsDay[i + 2],
        model_variables.Outlet2DS_Mult[i + 2 - model_variables.PlsDay[i + 2]],
        model_variables.Outlet2DS_Mult[i + 2],
        config["opt_qreg_mult"],
    )


def _calculate_outlet2dsrs(i: int, lo_model: pd.DataFrame, model_variables: object, lo_functions: object,
                           data: object, s80avg_l1: float, s80avg_l2: float, s80avg_l3: float, config: dict) -> None:
    """
    Calculate the Outlet2DSRS for the given index.

    Args:
        i (int): The current index.
        lo_model (pd.DataFrame): The model DataFrame.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        data (object): The data object.
        s80avg_l1 (float): The S80 average level 1.
        s80avg_l2 (float): The S80 average level 2.
        s80avg_l3 (float): The S80 average level 3.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Outlet2DSRS[i + 2] = lo_functions.Outlet2DSRS(
        model_variables.Release_Level[i + 2],
        data.S80_RegRelRates.at[0, "Zone_D1"],
        s80avg_l1,
        data.Pulses.at[
            (
                model_variables.PlsDay[i + 2] - 1
                if model_variables.PlsDay[i + 2] - 1 >= 0
                else len(data.Pulses) - 1
            ),
            f'S-80_L1_{config["schedule"]}',
        ],
        model_variables.Outlet2DS_Mult_2[i + 2],
        data.CE_SLE_turns.at[
            lo_model.at[i + 2, "date"].year - config["start_year"],
            "SLEturn",
        ],
        data.S80_RegRelRates.at[0, "Zone_D2"],
        s80avg_l2,
        data.Pulses.at[
            (
                model_variables.PlsDay[i + 2] - 1
                if model_variables.PlsDay[i + 2] - 1 >= 0
                else len(data.Pulses) - 1
            ),
            f'S-80_L2_{config["schedule"]}',
        ],
        data.S80_RegRelRates.at[0, "Zone_D3"],
        s80avg_l3,
        data.Pulses.at[
            (
                model_variables.PlsDay[i + 2] - 1
                if model_variables.PlsDay[i + 2] - 1 >= 0
                else len(data.Pulses) - 1
            ),
            f'S-80_L3_{config["schedule"]}',
        ],
        data.S80_RegRelRates.at[0, "Zone_C"],
        data.S80_RegRelRates.at[0, "Zone_B"],
        data.S80_RegRelRates.at[0, "Zone_A"],
    )
    diff = model_variables.Outlet2DSRS[i + 2] - lo_model.at[i + 2, "C44RO"]
    model_variables.Outlet2USRG1[i + 2] = max(
        0, diff.values[0] if isinstance(diff, pd.Series) else diff
    )
    model_variables.Sum_Outlet2USRG1[i + 2] = lo_functions.Sum_Outlet2USRG1(
        lo_model.at[i + 2, "date"].day, model_variables.Outlet2USRG1[i + 2]
    )


def _calculate_outlet2dsbs(i: int, model_variables: object, lo_functions: object, vlookup1_c: list,
                           outlet2_baseflow: float, config: dict) -> None:
    """
    Calculate the Outlet2DSBS for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        vlookup1_c (list): The vlookup1_c list.
        outlet2_baseflow (float): The outlet2 baseflow.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Outlet2DSBS[i + 2] = lo_functions.Outlet2DSBS(
        model_variables.Release_Level[i + 2],
        model_variables.Sum_Outlet2USRG1[i + 2],
        vlookup1_c[i],
        outlet2_baseflow,
        config["option_s80_baseflow"],
    )


def _calculate_outlet2usbk(i: int, model_variables: object, lo_functions: object, df_WSMs: pd.DataFrame,
                           lo_model: pd.DataFrame, data: object, config: dict) -> None:
    """
    Calculate the Outlet2USBK for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        df_WSMs (pd.DataFrame): The WSMs DataFrame.
        lo_model (pd.DataFrame): The model DataFrame.
        data (object): The data object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Outlet2USBK[i + 2] = lo_functions.Outlet2USBK(
        model_variables.Lake_Stage[i + 1],
        df_WSMs.at[i + 1, "D1"],
        model_variables.Outlet2USRG[i + 1],
        lo_model.at[i + 2, "C44RO"],
        data.SFWMM_Daily_Outputs.at[i + 2, "S308BK"],
        config["opt_s308"],
        config["s308_bk_const"],
        config["s308_bk_thr"],
    )
    model_variables.ROeast[i + 2] = (
        lo_model.at[i + 2, "C44RO"] - model_variables.Outlet2USBK[i + 2]
    )
    model_variables.Outlet2USBS[i + 2] = lo_functions.Outlet2USBS(
        model_variables.Outlet2DSBS[i + 2],
        model_variables.Outlet2USRG1[i + 2],
        model_variables.ROeast[i + 2],
        config["option_s80_baseflow"],
    )
    model_variables.Sum_Outlet2USBK[i + 2] = lo_functions.Sum_Outlet2USBK(
        lo_model.at[i + 2, "date"].day, model_variables.Outlet2USBK[i + 2]
    )


def _calculate_outlet2usrg_code(i: int, model_variables: object, lo_functions: object, data: object, config: dict,
                                date_range_6: pd.date_range) -> None:
    """
    Calculate the Outlet2USRG code for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        data (object): The data object.
        config (dict): The configuration dictionary.
        date_range_6 (pd.date_range): The date range.

    Returns:
        None
    """
    model_variables.Outlet2USRG_Code[i + 2] = lo_functions.Outlet2USRG_Code(
        model_variables.Outlet2USRG1[i + 2],
        model_variables.Outlet2USBS[i + 2],
        data.SFWMM_Daily_Outputs.at[i + 2, "S308RG"],
        data.SFWMM_Daily_Outputs.at[i + 2, "STEST"],
        config["option_reg_s77_s308"],
    )
    if config["sim_type"] == 0:
        model_variables.Outlet2USRG[i + 2] = lo_functions.Outlet2USRG(
            model_variables.Outlet2USRG_Code[i + 2],
            data.SFWMM_Daily_Outputs.at[i + 2, "S308RG"],
            data.SFWMM_Daily_Outputs.at[i + 2, "STEST"],
            config["opt_s308"],
            config["s308_rg_const"],
        )
    elif config["sim_type"] == 1:
        if model_variables.Lake_Stage[i + 1] >= 18:
            model_variables.Outlet2USRG[i + 2] = 7200
        elif model_variables.Lake_Stage[i + 1] <= 8:
            model_variables.Outlet2USRG[i + 2] = 0
        elif (tp_lake_s[i] <= p1) and (
            date_range_6[i + 2].month in [1, 2, 3, 4, 11, 12]
        ):
            model_variables.Outlet2USRG[i + 2] = s308_dv[
                (date_range_6[i + 2].month) - 1
            ]
        elif (tp_lake_s[i] <= p2) and (
            date_range_6[i + 2].month in [5, 6, 7, 8, 9, 10]
        ):
            model_variables.Outlet2USRG[i + 2] = s308_dv[
                (date_range_6[i + 2].month) - 1
            ]
        else:
            model_variables.Outlet2USRG[i + 2] = 0
    elif config["sim_type"] == 3:
        if model_variables.Lake_Stage[i + 1] >= 18:
            model_variables.Outlet2USRG[i + 2] = 7200
        else:
            model_variables.Outlet2USRG[i + 2] = 0


def _calculate_outlet2ds(i: int, model_variables: object, lo_functions: object, data: object, config: dict) -> None:
    """
    Calculate the Outlet2DS for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        data (object): The data object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Outlet2DS[i + 2] = lo_functions.S80(
        model_variables.ROeast[i + 2],
        model_variables.Outlet2USRG[i + 2],
        data.SFWMM_Daily_Outputs.at[i + 2, "S80"],
        config["s80_const"],
    )


def _calculate_outlet1us_mult(i: int, model_variables: object, lo_functions: object, seasons: pd.DataFrame,
                              config: dict) -> None:
    """
    Calculate the Outlet1US_Mult for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        seasons (pd.DataFrame): The seasons DataFrame.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Outlet1US_Mult[i + 2] = lo_functions.Outlet1US_Mult(
        seasons.at[i, "Season"],
        seasons.at[i, "Month"],
        model_variables.dh_7days[i + 1],
        model_variables.ReLevelCode_1[i + 2],
        model_variables.Fraction_of_Zone_height[i + 1],
        model_variables.ReLevelCode_2[i + 2],
        model_variables.ReLevelCode_3_S77[i + 2],
        config["opt_qreg_mult"],
    )


def _calculate_outlet1us_mult_2(i: int, model_variables: object, lo_functions: object, lo_model: pd.DataFrame, config: dict) -> None:
    """
    Calculate the Outlet1US_Mult_2 for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        lo_model (pd.DataFrame): The model DataFrame.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Outlet1US_Mult_2[i + 2] = lo_functions.Outlet1US_Mult_2(
        lo_model.at[i + 2, "date"].month,
        lo_model.at[i + 2, "date"].day,
        model_variables.PlsDay[i + 2],
        model_variables.Outlet1US_Mult[i + 2 - model_variables.PlsDay[i + 2]],
        model_variables.Outlet1US_Mult[i + 2],
        config["opt_qreg_mult"],
    )


def _calculate_outlet1usrs(i: int, model_variables: object, lo_functions: object, lo_model: pd.DataFrame, data: object,
                           config: dict, s77avg_l1: float, s77avg_l2: float, s77avg_l3: float) -> None:
    """
    Calculate the Outlet1USRS for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        lo_model (pd.DataFrame): The model DataFrame.
        data (object): The data object.
        config (dict): The configuration dictionary.
        s77avg_l1 (float): The S77 average level 1.
        s77avg_l2 (float): The S77 average level 2.
        s77avg_l3 (float): The S77 average level 3.

    Returns:
        None
    """
    model_variables.Outlet1USRS[i + 2] = lo_functions.Outlet1USRS(
        model_variables.Release_Level[i + 2],
        data.S77_RegRelRates.at[0, "Zone_D1"],
        s77avg_l1,
        data.Pulses.at[
            (
                model_variables.PlsDay[i + 2] - 1
                if model_variables.PlsDay[i + 2] - 1 >= 0
                else len(data.Pulses) - 1
            ),
            f'S-77_L1_{config["schedule"]}',
        ],
        model_variables.Outlet1US_Mult_2[i + 2],
        lo_model.at[i + 2, "C43RO"],
        data.CE_SLE_turns.at[
            lo_model.at[i + 2, "date"].year - config["start_year"],
            "CEturn",
        ],
        data.S77_RegRelRates.at[0, "Zone_D2"],
        s77avg_l2,
        data.Pulses.at[
            (
                model_variables.PlsDay[i + 2] - 1
                if model_variables.PlsDay[i + 2] - 1 >= 0
                else len(data.Pulses) - 1
            ),
            f'S-77_L2_{config["schedule"]}',
        ],
        model_variables.Zone_Code[i + 1],
        data.S77_RegRelRates.at[0, "Zone_D3"],
        s77avg_l3,
        data.Pulses.at[
            (
                model_variables.PlsDay[i + 2] - 1
                if model_variables.PlsDay[i + 2] - 1 >= 0
                else len(data.Pulses) - 1
            ),
            f'S-77_L3_{config["schedule"]}',
        ],
        data.S77_RegRelRates.at[0, "Zone_C"],
        data.S77_RegRelRates.at[0, "Zone_B"],
        data.S77_RegRelRates.at[0, "Zone_A"],
        config["opt_outlet1_dsrg"],
    )


def _calculate_sum_outlet1usrs(i: int, model_variables: object, lo_functions: object, lo_model: pd.DataFrame) -> None:
    """
    Calculate the Sum_Outlet1USRS for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        lo_model (pd.DataFrame): The model DataFrame.

    Returns:
        None
    """
    model_variables.Sum_Outlet1USRS[i + 2] = lo_functions.Sum_Outlet1USRS(
        lo_model.at[i + 2, "date"].day, model_variables.Outlet1USRS[i + 2]
    )


def _calculate_outlet1usbk(i: int, model_variables: object, lo_functions: object, lo_model: pd.DataFrame, data: object,
                           config: dict) -> None:
    """
    Calculate the Outlet1USBK for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        lo_model (pd.DataFrame): The model DataFrame.
        data (object): The data object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Outlet1USBK[i + 2] = lo_functions.Outlet1USBK(
        model_variables.Lake_Stage[i + 1],
        model_variables.Outlet1USRS[i + 2],
        model_variables.Outlet1USBSAP[i + 1],
        model_variables.Outlet1USEWS[i + 1],
        lo_model.at[i + 2, "C43RO"],
        data.SFWMM_Daily_Outputs.at[i + 2, "S77BK"],
        config["outlet1_usbk_switch"],
        config["outlet1_usbk_threshold"],
    )


def _calculate_rowest(i: int, model_variables: object, lo_model: pd.DataFrame) -> None:
    """
    Calculate the ROwest for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_model (pd.DataFrame): The model DataFrame.

    Returns:
        None
    """
    model_variables.ROwest[i + 2] = (
        lo_model.at[i + 2, "C43RO"] - model_variables.Outlet1USBK[i + 2]
    )


def _calculate_outlet1dsbs(i: int, model_variables: object, lo_functions: object, vlookup2_c: list,
                           outlet1_baseflow: float, config: dict) -> None:
    """
    Calculate the Outlet1DSBS for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        vlookup2_c (list): The vlookup2_c list.
        outlet1_baseflow (float): The outlet1 baseflow.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Outlet1DSBS[i + 2] = lo_functions.Outlet1DSBS(
        model_variables.Release_Level[i + 2],
        model_variables.Sum_Outlet1USRS[i + 2],
        vlookup2_c[i],
        outlet1_baseflow,
        config["option_s77_baseflow"],
    )


def _calculate_outlet1usbs(i: int, model_variables: object, lo_functions: object, config: dict) -> None:
    """
    Calculate the Outlet1USBS for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Outlet1USBS[i + 2] = lo_functions.Outlet1USBS(
        model_variables.Outlet1DSBS[i + 2],
        model_variables.Outlet1USRS[i + 2],
        model_variables.ROwest[i + 2],
        config["option_s77_baseflow"],
    )


def _calculate_outlet1usbsap(i: int, model_variables: object, lo_functions: object, config: dict) -> None:
    """
    Calculate the Outlet1USBSAP for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Outlet1USBSAP[i + 2] = lo_functions.Outlet1USBSAP(
        model_variables.Outlet1USBS[i + 2],
        model_variables.Post_Ap_Baseflow[i],
        config["opt_adap_prot"],
    )


def _calculate_outlet1usews(i: int, model_variables: object, lo_functions: object, data: object, config: dict) -> None:
    """
    Calculate the Outlet1USEWS for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        data (object): The data object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Outlet1USEWS[i + 2] = lo_functions.Outlet1USEWS(
        model_variables.Post_AP_EWS[i],
        data.SFWMM_Daily_Outputs.at[i + 2, "CAEST"],
        config["outlet1_usews_switch"],
        config["opt_adap_prot"],
    )


def _define_thc_class_normal_or_above(i: int, n_rows: int, model_variables: object, config: dict,
                                      adaptive_protocol_df: pd.DataFrame, df_WSMs: pd.DataFrame,
                                      targ_stg_df: pd.DataFrame, data: object, choose_1: float) -> None:
    """
    Define THC Class Normal or above for the given index.

    Args:
        i (int): The current index.
        n_rows (int): The number of rows.
        model_variables (object): The model variables object.
        config (dict): The configuration dictionary.
        adaptive_protocol_df (pd.DataFrame): The adaptive protocol DataFrame.
        df_WSMs (pd.DataFrame): The WSMs DataFrame.
        targ_stg_df (pd.DataFrame): The target stage DataFrame.
        data (object): The data object.
        choose_1 (float): The choose_1 value.

    Returns:
        None
    """
    if i < (n_rows - 2):
        thc_class_result = thc_class.THC_Class(
            config,
            i,
            model_variables.THC_Class_normal_or_above,
            model_variables.Lake_O_Stage_AP,
            model_variables.Lake_O_Schedule_Zone,
            model_variables.LStgCorres,
            model_variables.LowChance_Check,
            model_variables.Outlet1USRS_AP,
            model_variables.Outlet1USBS_AP,
            model_variables.Outlet1USRS_Pre_AP_S77_Baseflow,
            model_variables.Forecast_D_Sal,
            model_variables.n30d_mavg,
            model_variables.n30davgForecast,
            model_variables.LORS08_bf_rel,
            model_variables.LDS_LC6_1,
            model_variables.S_O,
            model_variables.All_4,
            model_variables.Sabf,
            model_variables.Swbf,
            model_variables.Swbu,
            model_variables.All_4andStage,
            model_variables.All_4andStagein,
            model_variables.P_AP_BF_Stg,
            model_variables.Logic_test_1,
            model_variables.Post_Ap_Baseflow,
            model_variables.Outlet1USRSplusPreAPS77bsf,
            model_variables.AndEstNeedsLakeWater,
            model_variables.AndLowChance61stagelessth11,
            model_variables.ATHCnora,
            model_variables.Choose_PAPEWS_1,
            model_variables.Choose_PAPEWS_2,
            model_variables.Post_AP_EWS,
            model_variables.Post_AP_Baseflow_EWS_cfs,
            adaptive_protocol_df,
            model_variables.Lake_Stage,
            model_variables.Zone_Code,
            df_WSMs,
            targ_stg_df,
            model_variables.Outlet1USRS,
            model_variables.Outlet1USBS,
            data.Estuary_needs_water,
            choose_1,
            model_variables.WSM_Zone,
        )
        model_variables.Post_Ap_Baseflow[i] = thc_class_result["Post_Ap_Baseflow"]
        model_variables.Post_AP_EWS[i] = thc_class_result["Post_AP_EWS"]


def _calculate_outlet1usreg(i: int, model_variables: object, lo_functions: object, data: object, config: dict,
                            tp_lake_s: float, s77_dv: float, date_range_6: pd.DatetimeIndex,
                            lo_model: pd.DataFrame, sensitivity_analysis_params) -> None:
    """
    Calculate the Outlet1USREG for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        data (object): The data object.
        config (dict): The configuration dictionary.
        tp_lake_s (float): The tp_lake_s value.
        s77_dv (float): The s77_dv value.
        date_range_6 (pd.DatetimeIndex): The date range.
        lo_model (pd.DataFrame): The model DataFrame.
        sensitivity_analysis_params (list): Holds a release rate for each month of the year.

    Returns:
        None
    """
    if config["sim_type"] == 0:
        model_variables.Outlet1USREG[i + 2] = lo_functions.Outlet1USREG(
            model_variables.Outlet1USRS[i + 2],
            model_variables.Outlet1USBSAP[i + 2],
            data.SFWMM_Daily_Outputs.at[i + 2, "S77RG"],
            config["outlet1_usreg_switch"],
            config["option_reg_s77_s308"],
        )
    elif config["sim_type"] == 1:
        if model_variables.Lake_Stage[i + 1] >= 18:
            model_variables.Outlet1USREG[i + 2] = 7800
        elif model_variables.Lake_Stage[i + 1] <= 8:
            model_variables.Outlet1USREG[i + 2] = 0
        elif (tp_lake_s[i] <= p1) and (
            date_range_6[i + 2].month in [1, 2, 3, 4, 11, 12]
        ):
            model_variables.Outlet1USREG[i + 2] = s77_dv[
                (date_range_6[i + 2].month) - 1
            ]
        elif (tp_lake_s[i] <= p2) and (
            date_range_6[i + 2].month in [5, 6, 7, 8, 9, 10]
        ):
            model_variables.Outlet1USREG[i + 2] = s77_dv[
                (date_range_6[i + 2].month) - 1
            ]
        else:
            model_variables.Outlet1USREG[i + 2] = 0
    elif config["sim_type"] == 3:
        if model_variables.Lake_Stage[i + 1] >= 18:
            model_variables.Outlet1USREG[i + 2] = 7200
        elif model_variables.Lake_Stage[i + 1] <= 8:
            model_variables.Outlet1USREG[i + 2] = 0
        elif lo_model.at[i + 2, 'date'].month == 1:
            model_variables.Outlet1USREG[i + 2] = lo_functions.Outlet_Rel_Sim(model_variables.Release_Level[i + 2], sensitivity_analysis_params[0])
        elif lo_model.at[i + 2, 'date'].month == 2:
            model_variables.Outlet1USREG[i + 2] = lo_functions.Outlet_Rel_Sim(model_variables.Release_Level[i + 2], sensitivity_analysis_params[1])
        elif lo_model.at[i + 2, 'date'].month == 3:
            model_variables.Outlet1USREG[i + 2] = lo_functions.Outlet_Rel_Sim(model_variables.Release_Level[i + 2], sensitivity_analysis_params[2])
        elif lo_model.at[i + 2, 'date'].month == 4:
            model_variables.Outlet1USREG[i + 2] = lo_functions.Outlet_Rel_Sim(model_variables.Release_Level[i + 2], sensitivity_analysis_params[3])
        elif lo_model.at[i + 2, 'date'].month == 5:
            model_variables.Outlet1USREG[i + 2] = lo_functions.Outlet_Rel_Sim(model_variables.Release_Level[i + 2], sensitivity_analysis_params[4])
        elif lo_model.at[i + 2, 'date'].month == 6:
            model_variables.Outlet1USREG[i + 2] = lo_functions.Outlet_Rel_Sim(model_variables.Release_Level[i + 2], sensitivity_analysis_params[5])
        elif lo_model.at[i + 2, 'date'].month == 7:
            model_variables.Outlet1USREG[i + 2] = lo_functions.Outlet_Rel_Sim(model_variables.Release_Level[i + 2], sensitivity_analysis_params[6])
        elif lo_model.at[i + 2, 'date'].month == 8:
            model_variables.Outlet1USREG[i + 2] = lo_functions.Outlet_Rel_Sim(model_variables.Release_Level[i + 2], sensitivity_analysis_params[7])
        elif lo_model.at[i + 2, 'date'].month == 9:
            model_variables.Outlet1USREG[i + 2] = lo_functions.Outlet_Rel_Sim(model_variables.Release_Level[i + 2], sensitivity_analysis_params[8])
        elif lo_model.at[i + 2, 'date'].month == 10:
            model_variables.Outlet1USREG[i + 2] = lo_functions.Outlet_Rel_Sim(model_variables.Release_Level[i + 2], sensitivity_analysis_params[9])
        elif lo_model.at[i + 2, 'date'].month == 11:
            model_variables.Outlet1USREG[i + 2] = lo_functions.Outlet_Rel_Sim(model_variables.Release_Level[i + 2], sensitivity_analysis_params[10])
        elif lo_model.at[i + 2, 'date'].month == 12:
            model_variables.Outlet1USREG[i + 2] = lo_functions.Outlet_Rel_Sim(model_variables.Release_Level[i + 2], sensitivity_analysis_params[11])


def _calculate_outlet1ds(i: int, model_variables: object, lo_functions: object, data: object, config: dict) -> None:
    """
    Calculate the Outlet1DS for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        data (object): The data object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Outlet1DS[i + 2] = lo_functions.Outlet1DS(
        model_variables.Outlet1USREG[i + 2],
        model_variables.Outlet1USEWS[i + 2],
        model_variables.ROwest[i + 2],
        data.SFWMM_Daily_Outputs.at[i + 2, "S79"],
        config["outlet1_ds_switch"],
    )


def _calculate_tot_reg_ew(i: int, model_variables: object) -> None:
    """
    Calculate the TotRegEW for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.

    Returns:
        None
    """
    model_variables.TotRegEW[i + 2] = (
        model_variables.Outlet1USREG[i + 2] + model_variables.Outlet2USRG[i + 2]
    ) * 1.9835


def _calculate_choose_wca(i: int, model_variables: object, lo_functions: object, data: object, config: dict) -> None:
    """
    Calculate the Choose_WCA and RegWCA for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        data (object): The data object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Choose_WCA[i + 2] = lo_functions.Choose_WCA(
        data.SFWMM_Daily_Outputs.at[i + 2, "RegWCA"],
        config["option_reg_wca"],
        config["constant_reg_wca"],
    )
    model_variables.RegWCA[i + 2] = min(
        config["max_cap_reg_wca"],
        config["multiplier_reg_wca"] * model_variables.Choose_WCA[i + 2],
    )


def _calculate_choose_l8c51(i: int, model_variables: object, lo_functions: object, data: object, config: dict) -> None:
    """
    Calculate the Choose_L8C51 and RegL8C51 for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        data (object): The data object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Choose_L8C51[i + 2] = lo_functions.Choose_L8C51(
        data.SFWMM_Daily_Outputs.at[i + 2, "RegL8C51"],
        config["option_reg_l8_c51"],
        config["constant_reg_l8_c51"],
    )
    model_variables.RegL8C51[i + 2] = min(
        config["max_cap_reg_l8_c51"],
        config["multiplier_reg_l8_c51"] * model_variables.Choose_L8C51[i + 2],
    )


def _calculate_tot_reg_so(i: int, model_variables: object) -> None:
    """
    Calculate the TotRegSo for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.

    Returns:
        None
    """
    model_variables.TotRegSo[i + 2] = (
        model_variables.RegWCA[i + 2] + model_variables.RegL8C51[i + 2]
    ) * 1.9835


def _calculate_stage2ar_and_stage2marsh(i: int, model_variables: object, stg_sto_ar: object) -> None:
    """
    Calculate the Stage2ar and Stage2marsh for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        stg_sto_ar (object): The stage storage area object.

    Returns:
        None
    """
    model_variables.Stage2ar[i + 2] = stg_sto_ar.stg2ar(model_variables.Lake_Stage[i + 1], 0)
    model_variables.Stage2marsh[i + 2] = stg_sto_ar.stg2mar(model_variables.Lake_Stage[i + 1], 0)


def _calculate_rf_and_et(i: int, model_variables: object, lo_functions: object, data: object, config: dict) -> None:
    """
    Calculate the RF and ET for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        data (object): The data object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.RF[i + 2] = data.RF_Vol.at[i + 2, "RFVol_acft"]
    model_variables.ET[i + 2] = lo_functions.ET(
        data.SFWMM_Daily_Outputs.at[i + 2, "et_dry"],
        model_variables.Stage2ar[i + 2],
        data.SFWMM_Daily_Outputs.at[i + 2, "et_litoral"],
        model_variables.Stage2marsh[i + 2],
        data.SFWMM_Daily_Outputs.at[i + 2, "et_open"],
        data.ET_Vol.at[i + 2, "ETVol_acft"],
        config["et_switch"],
    )


def _calculate_wsa(i: int, model_variables: object, lo_functions: object, df_WSMs: pd.DataFrame,
                   wca_stages_df: pd.DataFrame, tc_lonino_df: pd.DataFrame, data: object, config: dict) -> None:
    """
    Calculate the WSA for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        df_WSMs (pd.DataFrame): The WSMs DataFrame.
        wca_stages_df (pd.DataFrame): The WCA stages DataFrame.
        tc_lonino_df (pd.DataFrame): The TC Lonino DataFrame.
        data (object): The data object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Choose_WSA_1[i + 2] = lo_functions.Choose_WSA_1(
        df_WSMs.at[i + 2, "WSM1"],
        config["opt_wsa"],
        config["wsa_trig2"],
        config["wsa_off2"],
    )
    model_variables.Choose_WSA_2[i + 2] = lo_functions.Choose_WSA_2(
        df_WSMs.at[i + 2, "WSM1"],
        config["opt_wsa"],
        config["wsa_trig1"],
        config["wsa_off1"],
    )
    model_variables.WSA_MIA[i + 2] = lo_functions.WSA_MIA(
        wca_stages_df.at[i, "Are WCA stages too low?"],
        tc_lonino_df.at[i, "LONINO_Seasonal_Classes"],
        model_variables.Lake_Stage[i + 1],
        model_variables.Choose_WSA_1[i + 2],
        data.EAA_MIA_RUNOFF.at[i, "MIA"],
        data.EAA_MIA_RUNOFF.at[i, "S3PMP"],
        model_variables.Choose_WSA_2[i + 2],
        config["opt_wsa"],
        config["wsa_thc"],
        config["mia_cap2"],
        config["mia_cap1"],
    )
    model_variables.WSA_NNR[i + 2] = lo_functions.WSA_NNR(
        wca_stages_df.at[i, "Are WCA stages too low?"],
        tc_lonino_df.at[i, "LONINO_Seasonal_Classes"],
        model_variables.Lake_Stage[i + 1],
        model_variables.Choose_WSA_1[i + 2],
        data.EAA_MIA_RUNOFF.at[i, "NNR"],
        data.EAA_MIA_RUNOFF.at[i, "S2PMP"],
        model_variables.Choose_WSA_2[i + 2],
        config["opt_wsa"],
        config["wsa_thc"],
        config["nnr_cap2"],
        config["nnr_cap1"],
    )


def _calculate_dsto_and_storage(i: int, model_variables: object, lo_functions: object, storage_deviation: list,
                                start_storage: float) -> None:
    """
    Calculate the DSto and Storage for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        storage_deviation (list): The storage deviation list.
        start_storage (float): The start storage value.

    Returns:
        None
    """
    model_variables.DSto[i + 2] = (
        model_variables.NI_Supply[i + 2]
        + model_variables.RF[i + 2]
        - model_variables.ET[i + 2]
        + 1.9835
        * (
            model_variables.Outlet2USBK[i + 2]
            + model_variables.Outlet1USBK[i + 2]
            + model_variables.WSA_MIA[i + 2]
            + model_variables.WSA_NNR[i + 2]
            - model_variables.Outlet1USEWS[i + 2]
        )
        - model_variables.TotRegEW[i + 2]
        - model_variables.TotRegSo[i + 2]
        + storage_deviation[i + 2]
    )
    model_variables.Storage[i + 2] = lo_functions.Storage(
        model_variables.DayFlags[i + 2],
        model_variables.Storage[i],
        start_storage,
        model_variables.Storage[i + 1],
        model_variables.DSto[i + 2],
    )


def _calculate_lake_stage(i: int, model_variables: object, lo_functions: object, stg_sto_ar: object, data: object,
                          config: dict) -> None:
    """
    Calculate the Lake_Stage for the given index.

    Args:
        i (int): The current index.
        model_variables (object): The model variables object.
        lo_functions (object): The functions object.
        stg_sto_ar (object): The stage storage area object.
        data (object): The data object.
        config (dict): The configuration dictionary.

    Returns:
        None
    """
    model_variables.Lake_Stage[i + 2] = lo_functions.Lake_Stage(
        stg_sto_ar.stg2sto(model_variables.Storage[i + 2], 1),
        data.SFWMM_Daily_Outputs.at[i + 2, "EOD Stg(ft,NGVD)"],
        config["option_stage"],
    )


def _update_lo_model(lo_model: pd.DataFrame, model_variables: object) -> None:
    """
    Update the model DataFrame with the calculated model variables.

    Args:
        lo_model (pd.DataFrame): The model DataFrame.
        model_variables (object): The model variables object.

    Returns:
        None
    """
    lo_model["Cutback"] = model_variables.Cut_back
    lo_model["Stage"] = model_variables.Lake_Stage
    lo_model["Storage"] = model_variables.Storage
    lo_model["S77_Q"] = model_variables.Outlet1USREG
    lo_model["S308_Q"] = model_variables.Outlet2USRG
    lo_model["S77EW"] = model_variables.Outlet1USEWS
    lo_model["TotRegEW"] = model_variables.TotRegEW
    lo_model["TotRegSo"] = model_variables.TotRegSo


def _set_starting_storage(model_variables, start_storage):
    """
    Set the starting storage for the model variables.

    Args:
        model_variables (object): The model variables object.
        start_storage (float): The starting storage value.

    Returns:
        None
    """
    model_variables.Storage[0] = start_storage
    model_variables.Storage[1] = start_storage


def _determine_day_flags(startdate: pd.Timestamp, lo_model: pd.DataFrame, begdateCS: pd.Timestamp) -> str:
    """
    Determine the day flag based on the start date and model.

    Args:
        startdate (pd.Timestamp): The start date.
        lo_model (pd.DataFrame): The model DataFrame.
        begdateCS (pd.Timestamp): The beginning date for CS.

    Returns:
        str: The simulation day.
    """
    if (
        startdate.month == lo_model["date"].iloc[2].month
        and startdate.day == lo_model["date"].iloc[2].day
    ):
        return "SimDay1"
    elif (
        begdateCS.year == lo_model["date"].iloc[2].year
        and begdateCS.month == lo_model["date"].iloc[2].month
        and begdateCS.day == lo_model["date"].iloc[2].day
    ):
        return "CS start date"
    else:
        return lo_model["date"].iloc[2]


def _calculate_targ_stg_df(config: dict, data: object, date_range_5: pd.DatetimeIndex, model_variables: object) -> pd.DataFrame:
    """
    Calculate the target stage DataFrame.

    Args:
        config (dict): The configuration dictionary.
        data (object): The data object.
        date_range_5 (pd.DatetimeIndex): The date range.
        model_variables (object): The model variables object.

    Returns:
        pd.DataFrame: The target stage DataFrame.
    """
    targ_stg = data.Targ_Stg_June_1st if config["opt_date_targ_stg"] == 1 else data.Targ_Stg_May_1st

    targ_stg_df = pd.DataFrame(date_range_5, columns=["dates"])
    for i in range(len(targ_stg_df)):
        date = targ_stg_df["dates"].iloc[i]
        year = date.year
        day_of_year = date.timetuple().tm_yday

        model_variables.V10per[i] = replicate(year, day_of_year, 10, targ_stg)
        model_variables.V20per[i] = replicate(year, day_of_year, 20, targ_stg)
        model_variables.V25per[i] = replicate(year, day_of_year, 25, targ_stg)
        model_variables.V30per[i] = replicate(year, day_of_year, 30, targ_stg)
        model_variables.V40per[i] = replicate(year, day_of_year, 40, targ_stg)
        model_variables.V45per[i] = replicate(year, day_of_year, 45, targ_stg)
        model_variables.V50per[i] = replicate(year, day_of_year, 50, targ_stg)
        model_variables.V60per[i] = replicate(year, day_of_year, 60, targ_stg)

    targ_stg_df["10%"] = [x for x in model_variables.V10per if ~np.isnan(x)]
    targ_stg_df["20%"] = [x for x in model_variables.V20per if ~np.isnan(x)]
    targ_stg_df["25%"] = [x for x in model_variables.V25per if ~np.isnan(x)]
    targ_stg_df["30%"] = [x for x in model_variables.V30per if ~np.isnan(x)]
    targ_stg_df["40%"] = [x for x in model_variables.V40per if ~np.isnan(x)]
    targ_stg_df["45%"] = [x for x in model_variables.V45per if ~np.isnan(x)]
    targ_stg_df["50%"] = [x for x in model_variables.V50per if ~np.isnan(x)]
    targ_stg_df["60%"] = [x for x in model_variables.V60per if ~np.isnan(x)]

    return targ_stg_df


def _calculate_late_dry_season(adaptive_protocol_df: pd.DataFrame, tc_lonino_df: pd.DataFrame) -> None:
    """
    Calculate the Late Dry Season (Apr-May) logic and update the adaptive protocol DataFrame.

    Args:
        adaptive_protocol_df (pd.DataFrame): The adaptive protocol DataFrame.
        tc_lonino_df (pd.DataFrame): The TC Lonino DataFrame.

    Returns:
        None
    """
    late_dry_season = [3 < date.month < 6 for date in adaptive_protocol_df["date"]]
    adaptive_protocol_df["Late_Dry_Season"] = late_dry_season
    adaptive_protocol_df["Tributary Hydrologic Condition"] = tc_lonino_df["Tributary_Condition"]


def _calculate_basin_runoff(basin_ro: pd.DataFrame, data: object, outlet1_baseflow: float, outlet2_baseflow: float) -> None:
    """
    Calculate the basin runoff and update the basin runoff DataFrame.

    Args:
        basin_ro (pd.DataFrame): The basin runoff DataFrame.
        data (object): The data object.
        outlet1_baseflow (float): The outlet1 baseflow.
        outlet2_baseflow (float): The outlet2 baseflow.

    Returns:
        None
    """
    num_B_R = len(basin_ro.index)
    bs_c43ro = np.zeros(num_B_R)
    bs_c44ro = np.zeros(num_B_R)
    c44ro_sltrib = np.zeros(num_B_R)
    c44ro_bs = np.zeros(num_B_R)
    num_days = np.zeros(num_B_R)

    for i in range(num_B_R):
        num_days[i] = monthrange(
            basin_ro["date"].iloc[i].year, basin_ro["date"].iloc[i].month
        )[1]  # no. of days in each time step month.
        bs_c43ro[i] = max(0, (outlet1_baseflow - data.C43RO["C43RO"].iloc[i]))
        bs_c44ro[i] = max(0, (outlet2_baseflow - data.C44RO["C44RO"].iloc[i]))
        c44ro_sltrib[i] = bs_c44ro[i] + data.SLTRIB["SLTRIB_cfs"].iloc[i]
        c44ro_bs[i] = (
            max(0, data.C44RO["C44RO"].iloc[i] - outlet2_baseflow)
            * num_days[i]
        )

    basin_ro["Ndays"] = num_days
    basin_ro["C43RO"] = data.C43RO["C43RO"]
    basin_ro["BS-C43RO"] = bs_c43ro
    basin_ro["C44RO"] = data.C44RO["C44RO"]
    basin_ro["BS-C44RO"] = bs_c44ro
    basin_ro["SLTRIB"] = data.SLTRIB["SLTRIB_cfs"]
    basin_ro["C44RO_SLTRIB"] = c44ro_sltrib
    basin_ro["C44RO-BS"] = c44ro_bs


def _initialize_lo_model(date_range_6: pd.DatetimeIndex, data: object, config: dict) -> pd.DataFrame:
    """
    Initialize the model DataFrame with the given date range, data, and configuration.

    Args:
        date_range_6 (pd.DatetimeIndex): The date range.
        data (object): The data object.
        config (dict): The configuration dictionary.

    Returns:
        pd.DataFrame: The initialized LOONE model DataFrame.
    """
    lo_model = pd.DataFrame(date_range_6, columns=["date"])
    lo_model["Net_Inflow"] = data.NetInf_Input["Netflows_acft"]
    lo_model["LOSA_dmd_SFWMM"] = data.SFWMM_W_dmd["LOSA_dmd"] * (config["mult_losa"] / 100)
    lo_model["C44RO"] = data.C44_Runoff["C44RO"]
    return lo_model


def _determine_seasons(date_range_5: pd.DatetimeIndex, model_variables: object) -> pd.DataFrame:
    """
    Determine the seasons for each date in the given date range and update the model variables.

    Args:
        date_range_5 (pd.DatetimeIndex): The date range.
        model_variables (object): The model variables object.

    Returns:
        pd.DataFrame: The seasons DataFrame.
    """
    seasons = pd.DataFrame(date_range_5, columns=["date"])
    seasons_count = len(seasons.index)
    for i in range(seasons_count):
        month = seasons["date"].iloc[i].month
        if 2 < month < 6:
            S = 2
        elif 5 < month < 8:
            S = 3
        elif 7 < month < 11:
            S = 4
        else:
            S = 1
        model_variables.Daily_Seasons[i] = S
        model_variables.Mon[i] = month
    seasons["Season"] = model_variables.Daily_Seasons
    seasons["Month"] = model_variables.Mon
    return seasons


def _calculate_daily_water_demand(date_range_2: pd.DatetimeIndex, startdate: pd.Timestamp, data: object,
                                  config: dict) -> pd.DataFrame:
    """
    Calculate the daily water demand based on the given date range, start date, data, and configuration.

    Args:
        date_range_2 (pd.DatetimeIndex): The date range.
        startdate (pd.Timestamp): The start date.
        data (object): The data object.
        config (dict): The configuration dictionary.

    Returns:
        pd.DataFrame: The water demand DataFrame with daily demand.
    """
    water_demand = pd.DataFrame(date_range_2, columns=["date"])

    N = []
    week_number = []
    for i in water_demand["date"]:
        if i.month == startdate.month and i.day == startdate.day:
            n = 0
        else:
            n = n + 1
        N.append(n)
    water_demand["count"] = N

    for i in water_demand["count"]:
        if i > 363:
            J = 52
        else:
            J = int(i / 7) + 1
        week_number.append(J)
    water_demand["Week_num"] = week_number

    daily_demand = []
    for i in water_demand["Week_num"]:
        demand = ((data.Weekly_dmd[f'C{config["code"]}'].iloc[i - 1]) / 7) * (
            config["multiplier"] / 100
        )
        daily_demand.append(demand)
    water_demand["Daily_demand"] = daily_demand

    return water_demand


def _define_start_and_end_dates(config: dict) -> tuple:
    """
    Define the start and end dates based on the configuration.

    Args:
        config (dict): The configuration dictionary.

    Returns:
        tuple: A tuple containing the start date, begdateCS, and end date.
    """
    year, month, day = map(int, config["start_date_entry"])
    startdate = datetime(year, month, day).date()
    year, month, day = map(int, config["start_date_entry"])
    begdateCS = datetime(year, month, day).date()
    year, month, day = map(int, config["end_date_entry"])
    enddate = datetime(year, month, day).date()

    return startdate, begdateCS, enddate


def _create_dectree_df(date_range_5: pd.DatetimeIndex, tc_lonino_df: pd.DataFrame) -> pd.DataFrame:
    """
    Create the dectree DataFrame with the given date range and TC Lonino DataFrame.

    Args:
        date_range_5 (pd.DatetimeIndex): The date range.
        tc_lonino_df (pd.DataFrame): The TC Lonino DataFrame.

    Returns:
        pd.DataFrame: The decision tree DataFrame.
    """
    dec_tree_df = pd.DataFrame(date_range_5, columns=["Date"])
    dec_tree_df["Zone_B_MetFcast"] = tc_lonino_df["LONINO_Seasonal_Classes"]
    return dec_tree_df


def _calculate_pulse_averages(data: object, config: dict) -> dict:
    """
    Calculate the pulse averages for S-80 and S-77.

    Args:
        data (object): The data object containing pulse information.
        config (dict): The configuration dictionary.

    Returns:
        dict: A dictionary containing the pulse averages for S-80 and S-77.
    """
    pulse_averages = {
        "s80avg_l1": data.Pulses[f"S-80_L1_{config['schedule']}"].mean(),
        "s80avg_l2": data.Pulses[f"S-80_L2_{config['schedule']}"].mean(),
        "s80avg_l3": data.Pulses[f"S-80_L3_{config['schedule']}"].mean(),
        "s77avg_l1": data.Pulses[f"S-77_L1_{config['schedule']}"].mean(),
        "s77avg_l2": data.Pulses[f"S-77_L2_{config['schedule']}"].mean(),
        "s77avg_l3": data.Pulses[f"S-77_L3_{config['schedule']}"].mean(),
    }
    return pulse_averages


def _initialize_model_variables_stage_levels_flags(model_variables: object, config: dict, startdate: pd.Timestamp,
                                                   lo_model: pd.DataFrame, begdateCS: pd.Timestamp) -> None:
    """
    Initialize the model variables with the given configuration and start date.

    Args:
        model_variables (object): The model variables object.
        config (dict): The configuration dictionary.
        startdate (pd.Timestamp): The start date.
        lo_model (pd.DataFrame): The LOONE model DataFrame.
        begdateCS (pd.Timestamp): The beginning date for CS.

    Returns:
        None
    """
    model_variables.Lake_Stage[0] = config["beg_stage_cs"]
    model_variables.Lake_Stage[1] = config["beg_stage_cs"]
    model_variables.DecTree_Relslevel[0] = np.nan
    model_variables.DecTree_Relslevel[1] = np.nan
    model_variables.DayFlags[2] = _determine_day_flags(startdate, lo_model, begdateCS)


def LOONE_Q(workspace: str, p1: float, p2: float, s77_dv: float, s308_dv: float, tp_lake_s: float, sensitivity_analysis_params: List[float] = []) -> None:
    """This function runs the LOONE Q module.

    Args:
        workspace (str): The path to the workspace directory.
        p1 (float): Parameter 1.
        p2 (float): Parameter 2.
        s77_dv (float): s77_dv value.
        s308_dv (float): s308_dv value.
        tp_lake_s (float): tp_lake_s value.
        sensitivity_analysis_params (List[float], optional): Holds a release rate for each month of the year. Required when sim_type is 3. Defaults to [].

    Returns:
        None
    """
    os.chdir(workspace)
    config = load_config(workspace)

    data = DClass(workspace)
    model_variables = MVarClass(config)
    print("LOONE Q Module is Running!")
    # Based on the defined Start and End year, month, and day on the
    # Pre_defined_Variables File, Startdate and enddate are defined.
    startdate, begdateCS, enddate = _define_start_and_end_dates(config)

    ###################################################################
    if config["sim_type"] in [0, 1, 3]:
        df_wsms.WSMs(workspace)

    df_WSMs = pd.read_csv("df_WSMs.csv")

    # The Following Code interpolates daily LOSA demand from weekly
    # data for 6 differnet datasets where the user defines the LOSA
    # demand that will be used based on a Code (1:6).
    # Set time frame for model run
    date_range_2 = pd.date_range(start=startdate, end=enddate, freq="D")
    water_demand = _calculate_daily_water_demand(date_range_2, startdate, data, config)

    ###################################################################
    # Determine Tributary Hydrologic Conditions
    tc_lonino_df = trib_hc.Trib_HC(workspace)
    # Determine WCA Stages
    wca_stages_df = WCA_Stages_Cls(workspace, tc_lonino_df)
    # A dataframe to determine eachday's season (Months 11,12,1,2 are
    # Season 1, Months 3,4,5 are season 2, Months 6,7 are season 3,
    # Months 8,9,10 are season 4 )
    date_range_5 = pd.date_range(start=startdate, end=enddate, freq="D")
    seasons = _determine_seasons(date_range_5, model_variables)

    ###################################################################
    # This following Script runs the main model daily simulations.
    date_range_6 = pd.date_range(
        start=startdate - timedelta(days=1),
        end=enddate,
        freq="D",
    )
    lo_model = _initialize_lo_model(date_range_6, data, config)
    n_rows = len(lo_model.index)

    ##################################
    dec_tree_df = _create_dectree_df(date_range_5, tc_lonino_df)

    # Create a dataframe that includes Monthly Mean Basin Runoff &
    # BaseFlow-Runoff & Runoff-Baseflow (cfs)
    date_range_11 = pd.date_range(start=startdate, end=enddate, freq="MS")
    date_range_11d = pd.date_range(start=startdate, end=enddate, freq="D")
    date_range_11d.name = "Date"
    basin_ro = pd.DataFrame(date_range_11, columns=["date"])
    # Baseflows
    outlet1_baseflow = data.S77_RegRelRates["Zone_D0"].iloc[0]
    outlet2_baseflow = data.S80_RegRelRates["Zone_D0"].iloc[0]
    _calculate_basin_runoff(basin_ro, data, outlet1_baseflow, outlet2_baseflow)
    lo_model["C43RO"] = data.C43RO_Daily["C43RO"]
    pulse_averages = _calculate_pulse_averages(data, config)
    s80avg_l1 = pulse_averages["s80avg_l1"]
    s80avg_l2 = pulse_averages["s80avg_l2"]
    s80avg_l3 = pulse_averages["s80avg_l3"]
    s77avg_l1 = pulse_averages["s77avg_l1"]
    s77avg_l2 = pulse_averages["s77avg_l2"]
    s77avg_l3 = pulse_averages["s77avg_l3"]
    basin_ro = basin_ro.set_index(["date"])
    basin_ro.index = pd.to_datetime(basin_ro.index)
    basin_ro_daily = basin_ro.reindex(date_range_11d, method="ffill")
    basin_ro = basin_ro.reset_index()
    vlookup1 = basin_ro_daily["BS-C44RO"]
    vlookup1_c = [x for x in vlookup1 if ~np.isnan(x)]
    ###################################################################
    # This following script contains the logic and calculations for
    # the proposed Lake Okeechobee Adaptive Protocol.
    adaptive_protocol_df = pd.DataFrame(date_range_5, columns=["date"])
    _calculate_late_dry_season(adaptive_protocol_df, tc_lonino_df)

    # Define "Low Chance" 6/1 stg<11'
    targ_stg_df = _calculate_targ_stg_df(config, data, date_range_5, model_variables)

    # Outlet1_baseflow = Data.S77_RegRelRates['Zone_D0'].iloc[0]
    outlet1_baseflow = 450  # cfs
    vlookup2 = basin_ro_daily["BS-C43RO"]
    vlookup2_c = [x for x in vlookup2 if ~np.isnan(x)]
    ###################################################################
    _initialize_model_variables_stage_levels_flags(model_variables, config, startdate, lo_model, begdateCS)
    start_storage = stg_sto_ar.stg2sto(config["start_stage"], 0)
    _set_starting_storage(model_variables, start_storage)
    # Flood = np.zeros(n_rows, dtype = object)
    ##Here, I will insert the Storage Deviaiton Values as Input!
    storage_deviation = data.Storage_dev_df["DS_dev"]
    # Create a Choose Function for AP Post Baseflow
    # if Pre_defined_Variables.Opt_AdapProt == 0:
    #     C = 450
    # elif Pre_defined_Variables.Opt_AdapProt == 1:
    #     C = Data.S77_RegRelRates['Zone_D0'].iloc[0]
    # Choose_1 = C
    choose_1 = 450  # cfs
    ###################################################################
    _calculate_initial_zone_code_and_lo_zone(model_variables, lo_functions, df_WSMs)
    for i in range(n_rows - 2):
        _calculate_wsm_zone(i, model_variables, lo_functions, df_WSMs)
        _calculate_max_supply(i, model_variables, lo_functions, water_demand, config)
        _calculate_losa_supply(i, model_variables, lo_functions, lo_model, config)
        _calculate_ni_supply(i, model_variables, lo_model)
        _calculate_cutback(i, model_variables, lo_model)
        _calculate_demand_not_supplied(i, model_variables, lo_model)
        _calculate_zone_code(i, model_variables, lo_functions, df_WSMs)
        _generate_zone_column(i, model_variables, lo_functions)
        _calculate_decision_tree_variables(i, model_variables, dec_tree_functions, tc_lonino_df, config, seasons,
                                           dec_tree_df, df_WSMs)
        _calculate_release_levels(i, startdate, lo_model, model_variables, lo_functions, config, tc_lonino_df)
        _calculate_zone_codes(i, model_variables, lo_functions, df_WSMs)
        _calculate_release_level_codes(i, model_variables, lo_functions, config)
        _calculate_outlet2ds_mult(i, lo_model, model_variables, lo_functions, seasons, config)
        _calculate_outlet2dsrs(i, lo_model, model_variables, lo_functions, data, s80avg_l1, s80avg_l2, s80avg_l3, config)
        _calculate_outlet2dsbs(i, model_variables, lo_functions, vlookup1_c, outlet2_baseflow, config)
        _calculate_outlet2usbk(i, model_variables, lo_functions, df_WSMs, lo_model, data, config)
        _calculate_outlet2usrg_code(i, model_variables, lo_functions, data, config, date_range_6)
        _calculate_outlet2ds(i, model_variables, lo_functions, data, config)
        _calculate_relevel_code_3_s77(i, model_variables, lo_functions, config)
        _calculate_outlet1us_mult(i, model_variables, lo_functions, seasons, config)
        _calculate_outlet1us_mult_2(i, model_variables, lo_functions, lo_model, config)
        _calculate_outlet1usrs(i, model_variables, lo_functions, lo_model, data, config, s77avg_l1, s77avg_l2, s77avg_l3)
        _calculate_sum_outlet1usrs(i, model_variables, lo_functions, lo_model)
        _calculate_outlet1usbk(i, model_variables, lo_functions, lo_model, data, config)
        _calculate_rowest(i, model_variables, lo_model)
        _calculate_outlet1dsbs(i, model_variables, lo_functions, vlookup2_c, outlet1_baseflow, config)
        _calculate_outlet1usbs(i, model_variables, lo_functions, config)
        _define_thc_class_normal_or_above(i, n_rows, model_variables, config, adaptive_protocol_df, df_WSMs,
                                          targ_stg_df, data, choose_1)
        _calculate_outlet1usbsap(i, model_variables, lo_functions, config)
        _calculate_outlet1usews(i, model_variables, lo_functions, data, config)
        _calculate_outlet1usreg(i, model_variables, lo_functions, data, config, tp_lake_s, s77_dv, date_range_6, lo_model, sensitivity_analysis_params)
        _calculate_outlet1ds(i, model_variables, lo_functions, data, config)
        _calculate_tot_reg_ew(i, model_variables)
        _calculate_choose_wca(i, model_variables, lo_functions, data, config)
        _calculate_choose_l8c51(i, model_variables, lo_functions, data, config)
        _calculate_tot_reg_so(i, model_variables)
        _calculate_stage2ar_and_stage2marsh(i, model_variables, stg_sto_ar)
        _calculate_rf_and_et(i, model_variables, lo_functions, data, config)
        _calculate_wsa(i, model_variables, lo_functions, df_WSMs, wca_stages_df, tc_lonino_df, data, config)
        _calculate_dsto_and_storage(i, model_variables, lo_functions, storage_deviation, start_storage)
        _calculate_lake_stage(i, model_variables, lo_functions, stg_sto_ar, data, config)
        _update_lo_model(lo_model, model_variables)

    # Write out the results to a file - Needed because
    # execute_loone.py calls this script as a subprocess and can't get
    # this data.
    lo_model.to_csv("LOONE_Q_Outputs.csv")

    # Add scenario data to output df
    df_stage = pd.DataFrame(data = model_variables.Lake_Stage, columns = ['Outputs'])
    df_caloosahatchee = pd.DataFrame(data = model_variables.Outlet1USREG, columns = ['Outputs'])
    df_saint_lucie = pd.DataFrame(data = model_variables.Outlet2USRG, columns = ['Outputs'])
    df_south = pd.DataFrame(data = model_variables.TotRegSo/1.9835, columns = ['Outputs'])
    df_out = pd.concat([df_stage, df_caloosahatchee, df_saint_lucie, df_south])
    
    return [lo_model, df_out.T.loc['Outputs']]


if __name__ == "__main__":
    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        "workspace",
        nargs=1,
        help="The path to the working directory.",
    )
    argparser.add_argument("--p1", nargs=1)
    argparser.add_argument("--p2", nargs=1)
    argparser.add_argument("--s77_dv", nargs=1)
    argparser.add_argument("--s308_dv", nargs=1)
    argparser.add_argument("--tp_lake_s", nargs=1)
    args = argparser.parse_args()
    workspace = args.workspace[0]
    p1 = args.p1[0] if args.p1 else 0
    p2 = args.p2[0] if args.p2 else 0
    s77_dv = args.s77_dv[0] if args.s77_dv else 0
    s308_dv = args.s308_dv[0] if args.s308_dv else 0
    tp_lake_s = args.tp_lake_s[0] if args.tp_lake_s else 0
    LOONE_Q(workspace, p1, p2, s77_dv, s308_dv, tp_lake_s)



# This Script Interpolates each Water Shortage Management (WSMs) and each Regulation Schedule Breakpoint Zone (D, C, B, and A).
import os
import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from scipy import interpolate
from loone.data import Data as DClass
from loone.utils import load_config


def WSMs(workspace: str):
    """Generate WSMs (Weather State Modifiers) based on the given workspace.

    Args:
        workspace (str): The path to the workspace.

    Returns:
        None

    Raises:
        None
    """

    os.chdir(workspace)
    config = load_config(workspace)
    data = DClass(workspace)
    year, month, day = map(int, config["start_date_entry"])
    start_date = datetime(year, month, day).date()
    year, month, day = map(int, config["start_date_entry"])
    year, month, day = map(int, config["end_date_entry"])
    # Config end date + 1 day
    end_date = datetime(year, month, day).date() + timedelta(days=1)

    date_rng_1 = pd.date_range(start=start_date, end=end_date, freq="D")
    # Create a data frame with a date column
    df_wsms = pd.DataFrame(date_rng_1, columns=["date"])
    # Generate an annual cumulative day count
    wsm_length = len(df_wsms.index)
    operational_zones = list(data.WSMs_RSBKs)
    operational_zones.remove("Date")
    operational_zones.remove("Day")
    for i in operational_zones:
        globals()[i] = np.zeros(wsm_length)
    WSM_Count = date_rng_1.strftime("%j")
    for i in range(wsm_length):
        for j in operational_zones:
            globals()[j][i] = interpolate.interp1d(
                data.WSMs_RSBKs["Day"],
                data.WSMs_RSBKs[j],
                kind="linear",
            )(WSM_Count[i])

    df_wsms["count"] = WSM_Count
    for j in operational_zones:
        df_wsms[j] = globals()[j]
    if config["opt_new_tree"] == 1:
        df_wsms["C-b"] = data.WSMs_RSBKs["C-b_NewTree"]
    else:
        df_wsms["C-b"] = data.WSMs_RSBKs["C-b_NoNewTree"]
    df_wsms.drop(["C-b_NewTree", "C-b_NoNewTree"], axis=1, inplace=True)

    # Add one row in top of the dataframe (i.e. Dec/31/previous year) where the values = values of the original first row in the dataframe!
    first_row = pd.DataFrame(
        {
            "WSM4": df_wsms["WSM4"].iloc[0],
            "WSM3": df_wsms["WSM3"].iloc[0],
            "WSM2": df_wsms["WSM2"].iloc[0],
            "WSM1": df_wsms["WSM1"].iloc[0],
            "D0": df_wsms["D0"].iloc[0],
            "D1": df_wsms["D1"].iloc[0],
            "D2": df_wsms["D2"].iloc[0],
            "D3": df_wsms["D3"].iloc[0],
            "C": df_wsms["C"].iloc[0],
            "B": df_wsms["B"].iloc[0],
            "A": df_wsms["A"].iloc[0],
            "C-b": df_wsms["C-b"].iloc[0],
        },
        index=[0],
    )
    df_wsms = pd.concat([first_row, df_wsms]).reset_index(drop=True)

    df_wsms.to_csv("df_WSMs.csv")



# To return storage or surface area given stage (i=0) else (i = any number) will return stage given storage or surface area.
import pandas as pd
from scipy import interpolate


def stg2sto(v: float, i: int) -> float:
    """
    Calculate the Stage_Storage relationship for interpolation.

    Args:
        v (float): The stage value.
        i (int): The type of interpolation.

    Returns:
        float: The storage given the stage value, or the stage given the storage.
    """
    stgsto_data = pd.read_csv("StgSto_data.csv")
    # NOTE: We Can use cubic interpolation instead of linear
    x = stgsto_data["Stage"]
    y = stgsto_data["Storage"]
    if i == 0:
        # return storage given stage
        return interpolate.interp1d(
            x, y, fill_value="extrapolate", kind="linear"
        )(v)
    else:
        # return stage given storage
        return interpolate.interp1d(
            y, x, fill_value="extrapolate", kind="linear"
        )(v)


# Calculate the Stage_Area relationship for interpolation!
def stg2ar(v: float, i: int) -> float:
    """
    Calculate the Stage_Surface Area relationship for interpolation.

    Args:
        v(float): The stage value.
        i(int): The type of interpolation.

    Returns:
        float: The surface area given the stage value, or the stage given the surface area.
    """
    stgar_data = pd.read_csv("Stgar_data.csv")
    # NOTE: We Can use cubic interpolation instead of linear
    x = stgar_data["Stage"]
    y = stgar_data["Surf_Area"]
    if i == 0:
        # return surface area given stage
        return interpolate.interp1d(
            x, y, fill_value="extrapolate", kind="linear"
        )(v)
    else:
        # return stage given surface area
        return interpolate.interp1d(
            y, x, fill_value="extrapolate", kind="linear"
        )(v)


# Calculate the Stage_MarshArea relationship for interpolation!
def stg2mar(v: float, i: int) -> float:
    """
    Calculate the Stage_MarshArea relationship for interpolation.

    Args:
        v(float): The stage value.
        i(int): The type of interpolation.

    Returns:
        float: The interpolated value.
    """
    stgmar_data = pd.read_csv("Stgmar_data.csv")
    # NOTE: We Can use cubic interpolation instead of linear
    x = stgmar_data["Stage"]
    y = stgmar_data["Marsh_Ar"]
    if i == 0:
        # return marsh area given stage
        return interpolate.interp1d(
            x, y, fill_value="extrapolate", kind="linear"
        )(v)
    else:
        # return stage given marsh area
        return interpolate.interp1d(
            y, x, fill_value="extrapolate", kind="linear"
        )(v)



import os
import pandas as pd
import numpy as np
import scipy.stats as stats
import argparse


def bias_correct_file(file_path_observed: str,
                      file_path_simulated: str,
                      station: str,
                      forecast_mode: bool = False) -> pd.DataFrame:
    """Bias correct the simulated total phosphorus values against a station in Lake Okeechobee.

    Args:
        file_path_observed: str: The path to the file containing the observed total phosphorus data
        file_path_simulated: str: The path to the file containing the simulated total phosphorus data
        station: str: The name of the station where the total phosphorus values were observed. One of: 'L001', 'L005', 'L008', 'L004', 'L006', 'L007', 'LZ40'
        forecast_mode (bool, optional): Whether to run the model in forecast mode. Defaults to False.

    Returns:
        pd.DataFrame: A dataframe containing the bias corrected total phosphorus values for the station.
    """
    # Station Regions
    # NOTE: Station L008 is considered to be in both the north and south regions of the lake
    north_stations = ['L001', 'L005', 'L008']
    south_stations = ['L004', 'L006', 'L007', 'L008', 'LZ40']

    # Validate the station
    station = station.upper()

    if station not in north_stations and station not in south_stations:
        raise ValueError(f"Station {station} is not a valid station. Please choose from the following stations: {north_stations + south_stations}")

    # Check that input files exist
    if not os.path.exists(file_path_observed):
        raise FileNotFoundError(f"File {file_path_observed} does not exist.")

    if not os.path.exists(file_path_simulated):
        raise FileNotFoundError(f"File {file_path_simulated} does not exist.")

    # Get the region of the station
    station_region = 'TP_Lake_N' if station in north_stations else 'TP_Lake_S'

    # Read simulated TP concentrations in Lake Okeechobee
    df_simulated_total_phosphorus = pd.read_csv(file_path_simulated)
    df_simulated_total_phosphorus['date'] = pd.to_datetime(df_simulated_total_phosphorus['Date'])

    # Read observations at given station
    df_station_observations = pd.read_csv(file_path_observed)
    df_station_observations['date'] = pd.to_datetime(df_station_observations['date'])

    # Station is located in both the north and south regions of the lake
    if station in north_stations and station in south_stations:
        # Run bias correction for both regions
        simulated_bias_corrected_quantile_entry_percentile_north = bias_correct_df(df_simulated_total_phosphorus,
                                                                                   df_station_observations,
                                                                                   station,
                                                                                   'TP_Lake_N',
                                                                                   forecast_mode)
        simulated_bias_corrected_quantile_entry_percentile_south = bias_correct_df(df_simulated_total_phosphorus,
                                                                                   df_station_observations,
                                                                                   station,
                                                                                   'TP_Lake_S',
                                                                                   forecast_mode)

        df_results = pd.merge(simulated_bias_corrected_quantile_entry_percentile_north,
                              simulated_bias_corrected_quantile_entry_percentile_south,
                              how='outer',
                              on='date')

        # Return the bias corrected values for both regions
        return df_results
    else:
        # Run bias correction for the region where the station is located
        simulated_bias_corrected_quantile_entry_percentile = bias_correct_df(df_simulated_total_phosphorus,
                                                                             df_station_observations,
                                                                             station,
                                                                             station_region,
                                                                             forecast_mode)

        # Return the bias corrected values
        return simulated_bias_corrected_quantile_entry_percentile


def bias_correct_df(df_simulated_total_phosphorus: pd.DataFrame,
                    df_station_observations: pd.DataFrame,
                    station: str,
                    station_region: str,
                    forecast_mode: bool = False) -> pd.DataFrame:
    """Bias correct the simulated total phosphorus values against a station in Lake Okeechobee.

    Args:
        df_simulated_total_phosphorus: pd.DataFrame: A dataframe containing the simulated total phosphorus values
        df_station_observations: pd.DataFrame: A dataframe containing the observed total phosphorus values
        station: str: The name of the station where the total phosphorus values were observed. One of: 'L001', 'L005', 'L008', 'L004', 'L006', 'L007', 'LZ40'
        station_region: str: The region of the lake where the station is located. One of: 'TP_Lake_N', 'TP_Lake_S'
        forecast_mode (bool, optional): Whether to run the model in forecast mode. Defaults to False.

    Returns:
        np.array: A numpy array containing the bias corrected total phosphorus values for the station.
    """
    # Create a dataframe for the TP simulations
    df_simulated_total_phosphorus_region = pd.DataFrame()
    df_simulated_total_phosphorus_region['date'] = df_simulated_total_phosphorus['date']
    df_simulated_total_phosphorus_region[station_region] = df_simulated_total_phosphorus[station_region]
    df_simulated_total_phosphorus_region = df_simulated_total_phosphorus_region.set_index(['date'])
    df_simulated_total_phosphorus_region.index = pd.to_datetime(df_simulated_total_phosphorus_region.index, unit='ns')
    df_simulated_total_phosphorus_monthly = df_simulated_total_phosphorus_region.resample('ME').mean()

    # Create a dataframe for the observed TP values
    df_station_observations_processed = pd.DataFrame()
    df_station_observations_processed['date'] = df_station_observations['date']
    df_station_observations_processed[f'{station}_Obs_mg/m3'] = df_station_observations[f'{station}_PHOSPHATE, TOTAL AS P_mg/L'] * 1000  # mg/m3
    df_station_observations_processed = df_station_observations_processed.set_index(['date'])
    df_station_observations_processed.index = pd.to_datetime(df_station_observations_processed.index, unit='ns')
    df_station_observations_monthly = df_station_observations_processed.resample('ME').mean()

    df_bias_correct = pd.merge(df_simulated_total_phosphorus_monthly, df_station_observations_monthly, how='left', on='date')
    df_bias_correct_no_nan = df_bias_correct.dropna()
    observed_values = df_bias_correct_no_nan[f'{station}_Obs_mg/m3'].values
    simulated_values = df_bias_correct_no_nan[station_region].values

    # Determine R2 for simulations related to observations at the station
    if simulated_values.size == 0 or observed_values.size == 0:
        print(f'No data available for bias correction calculation for station {station}_{station_region[-1]}.')
        return df_bias_correct_no_nan
    slope, intercept, r_value, p_value, std_err = stats.linregress(simulated_values, observed_values)

    # Calculate R-squared (coefficient of determination)
    r_squared = r_value**2
    print(f'R2_{station}_{station_region[-1]}:', r_squared)

    # Calculate the quantiles of the observed and simulated time series.
    obs_quantiles = np.percentile(observed_values, [25, 50, 75])

    # Calculate the average of values less than the 25th quantile
    average_below_25th = np.mean([value for value in observed_values if value <= obs_quantiles[0]])
    # Calculate the average of values less than the 50th quantile (median)
    average_below_50th = np.mean([value for value in observed_values if obs_quantiles[0] < value <= obs_quantiles[1]])
    # Calculate the average of values less than the 75th quantile
    average_below_75th = np.mean([value for value in observed_values if obs_quantiles[1] < value <= obs_quantiles[2]])
    # Calculate the average of values more than the 75th quantile
    average_above_75th = np.mean([value for value in observed_values if value > obs_quantiles[2]])

    # Average of Sim
    average_tp = simulated_values.mean()

    # Calculate the correction factor
    correction_factor_1 = average_below_25th / average_tp
    correction_factor_2 = average_below_50th / average_tp
    correction_factor_3 = average_below_75th / average_tp
    correction_factor_4 = average_above_75th / average_tp

    # Apply the correction factor
    sim_BC_Qnt_EntPer = np.zeros(len(observed_values))

    for i in range(len(observed_values)):
        if observed_values[i] <= obs_quantiles[0]:
            sim_BC_Qnt_EntPer[i] = simulated_values[i] * correction_factor_1
        elif observed_values[i] > obs_quantiles[0] and observed_values[i] <= obs_quantiles[1]:
            sim_BC_Qnt_EntPer[i] = simulated_values[i] * correction_factor_2
        elif observed_values[i] > obs_quantiles[1] and observed_values[i] <= obs_quantiles[2]:
            sim_BC_Qnt_EntPer[i] = simulated_values[i] * correction_factor_3
        elif observed_values[i] > obs_quantiles[2]:
            sim_BC_Qnt_EntPer[i] = simulated_values[i] * correction_factor_4
        else:
            sim_BC_Qnt_EntPer[i] = simulated_values[i]

    # Calculate R-squared (coefficient of determination)
    if sim_BC_Qnt_EntPer.size == 0 or observed_values.size == 0:
        print(f'No data available for bias correction calculation for station {station}_{station_region[-1]}.')
        return df_bias_correct_no_nan
    slope, intercept, r_value, p_value, std_err = stats.linregress(sim_BC_Qnt_EntPer, observed_values)
    r_squared = r_value**2
    print(f'R2_{station}_{station_region[-1]}_BC_Qnt:', r_squared)

    # Check that length of bias corrected array matches the length of the date column in the dataframe
    assert len(sim_BC_Qnt_EntPer) == len(df_bias_correct_no_nan.index), 'The length of the bias corrected array does not match the length of the date column in the dataframe.'

    # Return the bias corrected values as a dataframe
    return pd.DataFrame(sim_BC_Qnt_EntPer, index=df_bias_correct_no_nan.index, columns=[station_region])


def main(input_dir: str,
         output_dir: str,
         simulated_file: str = 'LOONE_Nut_Output.csv',
         forecast_mode: bool = False) -> None:
    """Bias correct the simulated total phosphorus values against each station in Lake Okeechobee.

    Expects the input files to be in the format 'water_quality_{station}_PHOSPHATE, TOTAL AS P.csv' and 'LOONE_Nut_Output.csv'.
    This function expects the input files to be named 'water_quality_{station}_PHOSPHATE, TOTAL AS P.csv' for the observed data
    and 'LOONE_Nut_Output.csv' for the simulated data. It applies bias correction to the simulated total phosphorus values. The corrected
    values are saved to the specified output directory under the file name 'LOONE_Nut_Output_{station}_corrected.csv'.

    Args:
        input_dir: str: The path to the directory containing the input files.
        output_dir: str: The path to the directory where the output files will be saved.
        simulated_file: str: The name of the file containing the simulated Nut Ouput data. Defaults to 'LOONE_Nut_Output.csv'.
        forecast_mode (bool, optional): Whether to run the model in forecast mode. Defaults to False.

    Returns:
        None
    """

    # Bias correct the simulated total phosphorus values against each station
    for station in ['L001', 'L005', 'L008', 'L004', 'L006', 'L007', 'LZ40']:
        # Get the file paths to the observed and simulated total phosphorus data
        file_path_observed = os.path.join(input_dir, f'water_quality_{station}_PHOSPHATE, TOTAL AS P.csv')
        file_path_simulated = os.path.join(input_dir, simulated_file)

        # Bias correct the simulated total phosphorus values against the station
        df_results = bias_correct_file(file_path_observed, file_path_simulated, station, forecast_mode)

        # Save the bias corrected values to a file
        df_results.to_csv(os.path.join(output_dir, f'LOONE_Nut_Output_{station}_corrected.csv'))


if __name__ == '__main__':
    # Parse the command line arguments
    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        'input_dir',
        help='The path to the directory containing the input files.',
        nargs=1
    )
    argparser.add_argument(
        'output_dir',
        help='The path to the directory where the output files will be saved.',
        nargs=1
    )
    argparser.add_argument(
        '--simulated_file',
        help='The name of the file containing the simulated Nut Ouput data.',
        default='LOONE_Nut_Output.csv',
        nargs=1
    )
    argparser.add_argument(
        "--forecast_mode",
        action="store_true",
        help="Flag to indicate that the bias correction is running in forecast mode.",
    )

    args = argparser.parse_args()

    input_dir = args.input_dir[0]
    output_dir = args.output_dir[0]
    simulated_file = args.simulated_file[0]
    forecast_mode = args.forecast_mode

    # Run the main function
    main(input_dir, output_dir, simulated_file, forecast_mode)



# Nitrogen Model Functions

# For NO, NH4
def f_T_alt1(T, T_opt, T_min, T_max):
    import math
    if T < T_opt:
        f_T = math.exp(-2.3 * ((T - T_opt) / (T_opt - T_min)) ** 2)
    elif T >= T_opt:
        f_T = math.exp(-2.3 * ((T - T_opt) / (T_max - T_opt)) ** 2)
    return f_T


# For Chla
def f_T__Chla_alt1(T, T_opt, T_min, T_max, month):
    import math
    if T < T_opt and month in (6, 7, 8, 9, 10):
        f_T = math.exp(-2.3 * ((T - T_opt) / (T_opt - T_min)) ** 2) * 1.2
    elif T < T_opt and month in (11, 12, 1, 2, 3, 4, 5):
        f_T = math.exp(-2.3 * ((T - T_opt) / (T_opt - T_min)) ** 2) * 0.8
    elif T >= T_opt and month in (6, 7, 8, 9, 10):
        f_T = math.exp(-2.3 * ((T - T_opt) / (T_max - T_opt)) ** 2) * 1.2
    elif T >= T_opt and month in (11, 12, 1, 2, 3, 4, 5):
        f_T = math.exp(-2.3 * ((T - T_opt) / (T_max - T_opt)) ** 2) * 0.8
    return f_T


def f_T_alt2(T, T_opt, KTg1, KTg2):
    import math
    if T <= T_opt:
        f_T = math.exp(-KTg1 * ((T - T_opt) ** 2))
    else:
        f_T = math.exp(-KTg2 * ((T_opt - T) ** 2))
    return f_T


def f_T_alt3(T, T_ref, T_min):
    if T <= T_min:
        f_T = 0
    else:
        f_T = (T - T_min) / (T_ref - T_min)
    return f_T


def f_T_alt4(T, T_opt, k1, k2):
    import math
    if T <= T_opt:
        f_T = math.exp(-k1 * (T - T_opt) ** 2)
    else:
        f_T = math.exp(-k2 * (T - T_opt) ** 2)
    return f_T


def f_L_alt1(TD, RAD, Kw, Kc, Chla, z, K1, K2):
    import math
    I_z = (27.25 / TD) * RAD * math.exp(-(Kw + Kc * Chla) * z)
    f_L = (I_z * (1 + 2 * math.sqrt(K1 / K2))) / (I_z + K1 + I_z ** 2 / K2)
    return f_L


def f_L_alt2(fd, Kw, Kc, Chla, z, Io, Im):
    import math
    Ke = Kw + Kc * Chla
    f_L = (2.72 * fd / (Ke * z)) * (math.exp(-Io * math.exp(-Ke * z) / Im) - math.exp(-Io / Im))
    return f_L


def f_L_alt4(fd, Kw, z, PAR, Io, K_LP):
    import math
    PAR_d = PAR * Io / fd
    f_L = (fd / (Kw * z)) * math.log1p((K_LP + PAR_d) / (K_LP + PAR_d * math.exp(-Kw * z)))
    return f_L


def Nit_Rate(Method, Tot_nit_R, kni0, kni, NH4, K_NH, DO, KDO, fox_min, DO_Cr_n, DO_Opt_n, a):
    if Method == 'Simple':
        Nit_R = Tot_nit_R * NH4
    elif Method == 'Opt1':
        Nit_R = kni0 + kni * (NH4 / (K_NH + NH4)) * (DO / (KDO + DO))
    elif Method == 'Opt2':
        Nit_R = kni0 + kni * NH4 * ((1 - fox_min) * ((DO - DO_Cr_n) / (DO_Opt_n - DO_Cr_n)) ** (10 ** a) + fox_min)
    return Nit_R


def Denit_Rate(Method, Tot_denit_R, kden0, kden, NOx, KN, DO, KDO, DO_Cr_d, DO_Opt_d):
    if Method == 'Simple':
        Denit_R = Tot_denit_R * NOx
    elif Method == 'Opt1':
        Denit_R = kden0 + kden * (NOx / (KN + NOx)) * (1 - (DO / (KDO + DO)))
    elif Method == 'Opt2':
        Denit_R = kden0 + kden * NOx * ((DO_Cr_d - DO) / (DO_Cr_d - DO_Opt_d))
    return Denit_R


def NOx_Uptake_Fn(G_max, f_T, f_L, f_P, f_N, K_NH, K_TN, YNOChla, NH4, NOx, Chla):
    Uptake = (G_max * f_T * min(f_L, f_P, f_N) * (1 - (NH4 / (K_NH + NH4))) * ((NH4 + NOx) / (K_TN + NH4 + NOx)) * YNOChla * Chla)
    return Uptake


def NO_Sed(S_NO, Area, NO_Temp_Adj, Vol):
    NO_Sed = (S_NO * Area * NO_Temp_Adj) / Vol
    return NO_Sed


def NOx_Dynamics_alt1(L_ext, Q_o, NOx, kn_an, k_deni_T, NH4, Vol, G_max, f_T, f_L, f_P, f_N, K_NH, K_TN, YNOChla, Chla, S_NO, Area, NO_Temp_Adj):
    NOx_Nxt = ((L_ext - Q_o * NOx + (714.6 * 1000 * 1000) + kn_an * Vol - k_deni_T * Vol - (G_max * f_T * min(f_L, f_P, f_N) * (1 - (NH4 / (K_NH + NH4))) * ((NH4 + NOx) / (K_TN + NH4 + NOx)) * YNOChla * Chla) * Vol + S_NO * Area * NO_Temp_Adj) / Vol) + NOx
    return NOx_Nxt if NOx_Nxt >= 0 else 0


def NOx_Dynamics_N(L_ext, Atm_dep_N, Q_NS, NO_N, kn_an, k_deni_T, NH4_N, Vol_N, G_max, f_T, f_L, f_P_N, f_N_N, K_NH, K_TN, YNOChla, Chla_N, S_NO, Area_N, NO_Temp_Adj):
    NOx_Nxt = ((L_ext - Q_NS * NO_N + Atm_dep_N + kn_an * Vol_N - k_deni_T * Vol_N - (G_max * f_T * min(f_L, f_P_N, f_N_N) * (1 - (NH4_N / (K_NH + NH4_N))) * ((NH4_N + NO_N) / (K_TN + NH4_N + NO_N)) * YNOChla * Chla_N) * Vol_N + S_NO * Area_N * NO_Temp_Adj) / Vol_N) + NO_N
    return NOx_Nxt if NOx_Nxt >= 0 else 0


def NOx_Dynamics_S(Atm_dep_S, Q_N2S, Q_o, NO_N, NO_S, kn_an, k_deni_T, NH4_S, Vol_S, G_max, f_T, f_L, f_P_S, f_N_S, K_NH, K_TN, YNOChla, Chla_S, S_NO, Area_S, NO_Temp_Adj):
    NOx_Nxt = ((Q_N2S * NO_N - Q_o * NO_S + Atm_dep_S + kn_an * Vol_S - k_deni_T * Vol_S - (G_max * f_T * min(f_L, f_P_S, f_N_S) * (1 - (NH4_S / (K_NH + NH4_S))) * ((NH4_S + NO_S) / (K_TN + NH4_S + NO_S)) * YNOChla * Chla_S) * Vol_S + S_NO * Area_S * NO_Temp_Adj) / Vol_S) + NO_S
    return NOx_Nxt if NOx_Nxt >= 0 else 0


def NOx_N_DiffEq(NO_N, t, L_ext, Atm_dep_N, Q_NS, kn_an, k_deni_T, NH4_N, Vol_N, G_max, f_T, f_L, f_P_N, f_N_N, K_NH, K_TN, YNOChla, Chla_N, S_NO, Area_N, NO_Temp_Adj):
    dNOxdt = ((L_ext - Q_NS * NO_N + Atm_dep_N + kn_an * Vol_N - k_deni_T * Vol_N - (G_max * f_T * min(f_L, f_P_N, f_N_N) * (1 - (NH4_N / (K_NH + NH4_N))) * ((NH4_N + NO_N) / (K_TN + NH4_N + NO_N)) * YNOChla * Chla_N) * Vol_N + S_NO * Area_N * NO_Temp_Adj) / Vol_N)
    return dNOxdt


def NH4_N_DiffEq(NH4_N, t, L_ext, Atm_dep_N, Q_NS, kn_an, NO_N, Vol, G_max, f_T, f_L, f_P, f_N, K_NH, K_TN, Chla_N, YNHChla, K_r_T, S_NH4, Area_N, NO_Temp_Adj):
    dNH4dt = ((L_ext - Q_NS * NH4_N + Atm_dep_N - kn_an * NH4_N * Vol - (G_max * f_T * min(f_L, f_P, f_N) * (1 - (NH4_N / (K_NH + NH4_N))) * ((NH4_N + NO_N) / (K_TN + NH4_N + NO_N)) * YNHChla * Chla_N) * Vol + YNHChla * K_r_T * Chla_N * Vol + S_NH4 * Area_N * NO_Temp_Adj) / Vol)
    return dNH4dt


def Chla_N_alt1(Load, Q_NS, Chla_N, vc, z, K_m_T, K_r_T, Grazing, G_max, f_T, f_L, f_P_N, f_N_N, Vol_N):
    Chla_Nxt = ((Load - Q_NS * Chla_N - vc * Chla_N * Vol_N / z - K_m_T * Chla_N * Vol_N - K_r_T * Chla_N * Vol_N - Grazing * Vol_N + G_max * f_T * min(f_L, f_P_N, f_N_N) * Chla_N * Vol_N) / Vol_N) + Chla_N
    return Chla_Nxt if Chla_Nxt >= 5 else 5


def Chla_N_alt2(Load, Q_NS, Chla_N, vc, z, K_m_T, K_r_T, Grazing, G_max, f_T, f_L, f_P_N, f_N_N, Vol_N):
    Chla_Nxt = ((Load - Q_NS * Chla_N - vc * Chla_N * Vol_N / z - K_m_T * Chla_N * Vol_N - K_r_T * Chla_N * Vol_N - Grazing * Vol_N + G_max * f_T * f_L * min(f_P_N, f_N_N) * Chla_N * Vol_N) / Vol_N) + Chla_N
    return Chla_Nxt if Chla_Nxt >= 5 else 5


def Chla_Growth_N(Chla_N, G_max, f_T, f_L, f_P_N, f_N_N, Vol_N):
    Growth = (G_max * f_T * f_L * min(f_P_N, f_N_N) * Chla_N * Vol_N) / Vol_N
    return Growth if Growth >= 5 else 5


def NOx_S_DiffEq(NO_S, t, Atm_dep_S, Q_N2S, Q_o, NO_N, kn_an, k_deni_T, NH4_S, Vol_S, G_max, f_T, f_L, f_P_S, f_N_S, K_NH, K_TN, YNOChla, Chla_S, S_NO, Area_S, NO_Temp_Adj):
    dNOxdt = ((Q_N2S * NO_N - Q_o * NO_S + Atm_dep_S + kn_an * Vol_S - k_deni_T * Vol_S - (G_max * f_T * min(f_L, f_P_S, f_N_S) * (1 - (NH4_S / (K_NH + NH4_S))) * ((NH4_S + NO_S) / (K_TN + NH4_S + NO_S)) * YNOChla * Chla_S) * Vol_S + S_NO * Area_S * NO_Temp_Adj) / Vol_S)
    return dNOxdt


def NH4_S_DiffEq(NH4_S, t, Atm_dep_S, Q_N2S, Q_o, NH4_N, kn_an, NO_S, Vol, G_max, f_T, f_L, f_P, f_N, K_NH, K_TN, Chla_S, YNHChla, K_r_T, S_NH4, Area_S, NO_Temp_Adj):
    dNH4dt = ((Q_N2S * NH4_N - Q_o * NH4_S + Atm_dep_S - kn_an * NH4_S * Vol - (G_max * f_T * min(f_L, f_P, f_N) * (1 - (NH4_S / (K_NH + NH4_S))) * ((NH4_S + NO_S) / (K_TN + NH4_S + NO_S)) * YNHChla * Chla_S) * Vol + YNHChla * K_r_T * Chla_S * Vol + + S_NH4 * Area_S * NO_Temp_Adj) / Vol)
    return dNH4dt


def Chla_S_alt1(Q_N2S, Q_o, Chla_N, Chla_S, vc, z, K_m_T, K_r_T, Grazing, G_max, f_T, f_L, f_P_S, f_N_S, Vol_S):
    Chla_Nxt = ((Q_N2S * Chla_N - Q_o * Chla_S - vc * Chla_S * Vol_S / z - K_m_T * Chla_S * Vol_S - K_r_T * Chla_S * Vol_S - Grazing * Vol_S + G_max * f_T * min(f_L, f_P_S, f_N_S) * Chla_S * Vol_S) / Vol_S) + Chla_S
    return Chla_Nxt if Chla_Nxt >= 5 else 5


def Chla_S_alt2(Q_N2S, Q_o, Chla_N, Chla_S, vc, z, K_m_T, K_r_T, Grazing, G_max, f_T, f_L, f_P_S, f_N_S, Vol_S):
    Chla_Nxt = ((Q_N2S * Chla_N - Q_o * Chla_S - vc * Chla_S * Vol_S / z - K_m_T * Chla_S * Vol_S - K_r_T * Chla_S * Vol_S - Grazing * Vol_S + G_max * f_T * f_L * min(f_P_S, f_N_S) * Chla_S * Vol_S) / Vol_S) + Chla_S
    return Chla_Nxt if Chla_Nxt >= 5 else 5


def Chla_Growth_S(Chla_S, G_max, f_T, f_L, f_P_S, f_N_S, Vol_S):
    Growth = (G_max * f_T * f_L * min(f_P_S, f_N_S) * Chla_S * Vol_S) / Vol_S
    return Growth if Growth >= 0 else 0


# def NOx_N_DiffEq(NO_N, t, L_ext, Atm_dep_N, Q_NS, kn_an, k_deni_T, NH4_N, Vol_N, G_max, f_T, f_L, f_P_N, f_N_N, K_NH, K_TN, YNOChla, Chla_N, S_NO, Area_N, NO_Temp_Adj):
#     dNOxdt = ((L_ext - Q_NS * NO_N + Atm_dep_N + kn_an * Vol_N - k_deni_T * Vol_N - (G_max * f_T * f_L * min(f_P_N, f_N_N) * (1 - (NH4_N / (K_NH + NH4_N))) * ((NH4_N + NO_N) / (K_TN + NH4_N + NO_N)) * YNOChla * Chla_N) * Vol_N + S_NO * Area_N * NO_Temp_Adj) / Vol_N)
#     return dNOxdt

# def NOx_S_DiffEq(NO_S, t, Atm_dep_S, Q_N2S, Q_o, NO_N, kn_an, k_deni_T, NH4_S, Vol_S, G_max, f_T, f_L, f_P_S, f_N_S, K_NH, K_TN, YNOChla, Chla_S, S_NO, Area_S, NO_Temp_Adj):
#     dNOxdt = ((Q_N2S * NO_N - Q_o * NO_S + Atm_dep_S + kn_an * Vol_S - k_deni_T * Vol_S - (G_max * f_T * f_L * min(f_P_S, f_N_S) * (1 - (NH4_S / (K_NH + NH4_S))) * ((NH4_S + NO_S) / (K_TN + NH4_S + NO_S)) * YNOChla * Chla_S) * Vol_S + S_NO * Area_S * NO_Temp_Adj) / Vol_S)
#     return dNOxdt


def NOx_Dynamics_alt2(L_ext, Q_o, NOx, kn_an, k_deni_T, NH4, Vol, G_max, f_T, f_L, f_P, f_N, K_NH, K_TN, YNOChla, Chla, S_NO, Area, NO_Temp_Adj):
    NOx_Nxt = ((L_ext - Q_o * NOx + (714.6 * 1000 * 1000) + kn_an * Vol - k_deni_T * Vol - (G_max * f_T * f_L * min(f_P, f_N) * (1 - (NH4 / (K_NH + NH4))) * ((NH4 + NOx) / (K_TN + NH4 + NOx)) * YNOChla * Chla) * Vol + S_NO * Area * NO_Temp_Adj) / Vol) + NOx
    return NOx_Nxt if NOx_Nxt >= 0 else 0


def NH4_Dynamics_v1(L_ext, Q_o, NOx, kn_an, NH4, Vol, G_max, f_T, f_L, f_P, f_N, K_NH, K_TN, YNOChla, Chla, YNHChla, K_r_T):
    NH4_Nxt = ((L_ext - Q_o * NH4 + (661.7 * 1000 * 1000) - kn_an * NH4 * Vol - (G_max * f_T * min(f_L, f_P, f_N) * (1 - (NH4 / (K_NH + NH4))) * ((NH4 + NOx) / (K_TN + NH4 + NOx)) * YNOChla * Chla) * Vol + YNHChla * K_r_T * Chla * Vol) / Vol) + NH4
    return NH4_Nxt if NH4_Nxt >= 0 else 0


def Chl_a_Dynamics_alt1(Load, Q_o, Chla, vc, z, K_m_T, K_r_T, Grazing, G_max, f_T, f_L, f_P, f_N, Vol):
    Chla_Nxt = ((Load - Q_o * Chla - vc * Chla * Vol / z - K_m_T * Chla * Vol - K_r_T * Chla * Vol - Grazing * Vol + G_max * f_T * min(f_L, f_P, f_N) * Chla * Vol) / Vol) + Chla
    return Chla_Nxt if Chla_Nxt >= 0 else 0


def Chl_a_Dynamics_alt2(Load, Q_o, Chla, vc, z, K_m_T, K_r_T, Grazing, G_max, f_T, f_L, f_P, f_N, Vol):
    Chla_Nxt = ((Load - Q_o * Chla - vc * Chla * Vol / z - K_m_T * Chla * Vol - K_r_T * Chla * Vol - Grazing * Vol + G_max * f_T * f_L * min(f_P, f_N) * Chla * Vol) / Vol) + Chla
    return Chla_Nxt if Chla_Nxt >= 0 else 0


def Chl_a_Dynamics_alt3(Load, Q_o, Chla, vc, z, K_m_T, K_r_T, Grazing, G_max, f_T, f_L, f_P, f_N, Vol):
    Chla_Nxt = ((Load - Q_o * Chla - vc * Chla * Vol / z - K_m_T * Chla * Vol - K_r_T * Chla * Vol - Grazing * Vol + G_max * f_T * f_L * min(f_P, f_N) * Chla * Vol) / Vol) + Chla
    return Chla_Nxt if Chla_Nxt >= 0 else 0


def Chl_a_Dynamics_alt4(Load, Q_o, Chla, vc, z, K_m_T, K_r_T, Grazing, G_max, f_T, f_L, f_P, f_N, Vol):
    Chla_Nxt = ((Load - Q_o * Chla - vc * Chla * Vol / z - K_m_T * Chla * Vol - K_r_T * Chla * Vol - Grazing * Vol + G_max * f_T * f_L * f_P * f_N * Chla * Vol) / Vol) + Chla
    return Chla_Nxt if Chla_Nxt >= 0 else 0
